/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "../config.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <gdk/gdkx.h>
#include <kiba-plugin.h>
#include <kiba-prefs.h>
#include <kiba-object.h>
#include <kiba-dock.h>
#include <kiba-screen.h>
#include <kiba-debug.h>
#include <kiba-desktop-icon.h>
#include <kiba-x11.h>

#include "launcher.h"

/**
Helper Functions
**/

static gint kiba_launcher_string_to_flag (const char *string) {
	static const char *launcher_flag_data[] = {
		"launcher_grouping", "launcher_control_class"
	};
	
	gint i;
	
	for (i = 0; i < G_N_ELEMENTS(launcher_flag_data); i++) {
		if (strcmp (launcher_flag_data[i], string) == 0)
			return i;
	}
	
	return KIBA_LAUNCHER_FLAG_NONE;
}

/**
KibaPrefs Callbacks
**/

static void dock_object_size_cb (const char *option_id,
			         gpointer    value,
			         gpointer    data) {
	KibaLauncher *launcher = data;
	KibaDesktopIcon *icon;
	GList *l;
	
	for (l = launcher->icons; l != NULL; l = l->next) {
		icon = l->data;
		
		icon->surface->width = icon->object->allocation.width;
		icon->surface->height = icon->object->allocation.height;
		kiba_desktop_icon_update_surface (icon);
	}
}

static void object_size_factor_cb (const char *option_id,
			           gpointer    value,
			           gpointer    data) {
	KibaLauncher *launcher = data;
	KibaLauncherIconGroup *group;
	KibaDesktopIcon *icon;
	KibaDock *dock;
	GList *l;
	
	dock = kiba_dock_get();
	launcher->size_scale = g_ascii_strtod((char*) value, NULL);
	
	for (l = launcher->groups; l != NULL; l = l->next) {
		group = l->data;
		
		kiba_object_set_requisition_factor (KIBA_OBJECT(group->widget),
			launcher->size_scale, launcher->size_scale);
	}
	
	for (l = launcher->icons; l != NULL; l = l->next) {
		icon = l->data;
		
		icon->surface->width = icon->object->allocation.width;
		icon->surface->height = icon->object->allocation.height;
		kiba_desktop_icon_update_surface (icon);
	}
	gtk_widget_queue_resize (GTK_WIDGET(dock));
}

static void launcher_object_alpha_cb (const char *option_id,
				      gpointer    value,
				      gpointer    data) {
	KibaLauncher *launcher = data;
	KibaLauncherIconGroup *group;
	GList *l;
	float alpha;
	
	alpha = g_ascii_strtod((char*) value, NULL);
	
	alpha = MAX(alpha, -alpha);
	alpha = MIN(alpha, 1.0);
	
	launcher->alpha = alpha;
	
	for (l = launcher->groups; l != NULL; l = l->next) {
		group = l->data;
		
		kiba_object_set_alpha (KIBA_OBJECT(group->widget), alpha);
	}
}

static void launcher_flags_cb (const char *option_id,
			       gpointer    value,
			       gpointer    data) {
	KibaLauncher *launcher = data;
	gint flag;
	
	flag = kiba_launcher_string_to_flag(option_id);
	if (flag == KIBA_LAUNCHER_FLAG_NONE)
		return;
	
	if ((gboolean) GPOINTER_TO_INT(value) == TRUE)
		kiba_launcher_set_flag (launcher, flag);
	else
		kiba_launcher_unset_flag (launcher, flag);
}

/**
Plugin Functions
**/

static void launcher_plugin_unload (KibaPlugin *plugin) {
	KibaLauncher *launcher;
	
	launcher = g_object_get_data (G_OBJECT(plugin), LAUNCHER_PLUGIN_NAME);
	
	kiba_prefs_remove_callback_by_data (launcher);
	
	kiba_launcher_destroy (launcher);
	
	g_object_set_data (G_OBJECT(plugin),
		LAUNCHER_PLUGIN_NAME, NULL);
}

static gboolean launcher_plugin_load (KibaPlugin *plugin) {
	KibaLauncher *launcher;
	GError *error = NULL;
	const char *homedir;
	char *file_name;
	GDir *dir;
	
#ifdef ENABLE_NLS
	bindtextdomain (GETTEXT_PACKAGE, DATADIR"/locale");
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
#endif
	
	homedir = g_get_home_dir ();
	
	char launcher_dir[strlen(homedir)+strlen(LAUNCHER_DIR)];
	sprintf (launcher_dir, "%s" LAUNCHER_DIR, homedir);
	if (!g_file_test (launcher_dir, G_FILE_TEST_IS_DIR)) {
		if (g_mkdir (launcher_dir, 7664) != 0) {
			DEBUG_ERROR ("Failed to Create %s\n", launcher_dir);
			return FALSE;
		}
	}
	
	launcher = kiba_launcher_new(G_TYPE_MODULE(plugin));
	if (!KIBA_IS_LAUNCHER(launcher)) {
		DEBUG_ERROR ("Failed to create the KibaLauncher Instance\n");
		return FALSE;
	}
	
	g_object_set_data (G_OBJECT(plugin),
		LAUNCHER_PLUGIN_NAME, (gpointer) launcher);
	
	kiba_prefs_add_callback ("launcher_grouping",
		launcher_flags_cb, launcher, TRUE);
	kiba_prefs_add_callback ("launcher_alpha",
		launcher_object_alpha_cb, launcher, TRUE);
	kiba_prefs_add_callback ("launcher_size_factor",
		object_size_factor_cb, launcher, TRUE);
	kiba_prefs_add_callback ("dock_object_size",
		dock_object_size_cb, launcher, TRUE);
	
	/* add desktop files from homedir */
	dir = g_dir_open (launcher_dir, 0, &error);
	if (error != NULL) {
		DEBUG_ERROR ("Failed to open '%s': %s\n",
			launcher_dir, error->message);
		g_error_free (error);
		return FALSE;
	}
	
	while ((file_name = (char*) g_dir_read_name (dir)) != NULL) {
		KibaDesktopIcon *icon;
		
		if (!g_str_has_suffix (file_name, ".desktop"))
			continue;
		
		char uri[strlen(launcher_dir) + strlen(file_name) + 1];
		sprintf (uri, "%s/%s", launcher_dir, file_name);
		
		icon = kiba_desktop_icon_new (G_TYPE_MODULE(plugin));
		if (kiba_desktop_icon_set_file (icon, uri))
			kiba_launcher_add (launcher, icon);
		else
			kiba_desktop_icon_destroy (icon);
	}
	g_dir_close (dir);
	
	return TRUE;
}

static PrefMember launcher_options[] = {
  { "launcher_alpha", "1.0", KIBA_DOUBLE },
  { "launcher_size_factor", "1.0", KIBA_DOUBLE },
  { "launcher_grouping", "FALSE", KIBA_BOOLEAN }
};

static KibaPluginVtable vtable = {
	LAUNCHER_PLUGIN_NAME,		/* id/name        */
	"Launcher Plugin for Kiba-Dock",/* summary        */
	NULL,				/* description    */	

	"Daniel Bessler <danielb@kiba-dock.org>",
					/* author         */
	"http://www.kiba-dock.org",	/* homepage       */

	launcher_plugin_load,		/* load            */
	launcher_plugin_unload,		/* unload          */
};

KIBA_INIT_PLUGIN (vtable, launcher_options);
