/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "../config.h"
#include <string.h>
#include <gdk/gdkx.h>
#include <X11/Xlib.h>
#include <libintl.h>

#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-plugin.h>
#include <kiba-icon.h>
#include <kiba-object.h>
#include <kiba-menu-items.h>
#include <kiba-info-win.h>
#include <kiba-icon-view.h>
#include <kiba-image.h>
#include <kiba-app-chooser.h>
#include <kiba-desktop-editor.h>
#include <kiba-screen.h>
#include <kiba-x11.h>
#include <kiba-key-file.h>

#include "launcher.h"

#define _(string)  dgettext (GETTEXT_PACKAGE, string)

enum {
	ICON_ADDED_EVENT = 0,
	ICON_REMOVED_EVENT,
	SIG_KIBA_LAUNCHER_LAST
};

static guint kiba_launcher_signals[SIG_KIBA_LAUNCHER_LAST] = { 0 };

static void kiba_launcher_create_group_for_icon (KibaLauncher    *launcher,
						 KibaDesktopIcon *icon);
static gboolean kiba_launcher_select_icon_cb (GtkWidget *widget,
					        gpointer   data);
static void kiba_launcher_icon_view_size_allocate_cb (GtkWidget     *widget,
						      GtkAllocation *allocation,
						      gpointer       data);

static void     kiba_launcher_init              (KibaLauncher      *self);
static void     kiba_launcher_class_init        (KibaLauncherClass *klass);

static gpointer kiba_launcher_parent_class = NULL;
static GType    kiba_launcher_type_id = 0;

static void     kiba_launcher_class_intern_init (gpointer klass)
{
  kiba_launcher_parent_class = g_type_class_peek_parent (klass); 
  kiba_launcher_class_init ((KibaLauncherClass*) klass); 
}

GType
kiba_launcher_get_type (void)
{
  return kiba_launcher_type_id;
}

static void
kiba_launcher_register_type (GTypeModule *type_module)
{
  const GTypeInfo g_define_type_info = {
    sizeof (KibaLauncherClass),
    (GBaseInitFunc) NULL,
    (GBaseFinalizeFunc) NULL,
    (GClassInitFunc) kiba_launcher_class_intern_init,
    (GClassFinalizeFunc) NULL,
    NULL,   /* class_data */
    sizeof (KibaLauncher),
    0,      /* n_preallocs */
    (GInstanceInitFunc) kiba_launcher_init, 
    NULL    /* value_table */
  };
  kiba_launcher_type_id = g_type_module_register_type (type_module,
					      G_TYPE_OBJECT,
					      "KibaLauncher",
					      &g_define_type_info, 0);
}

/**
Helper Functions
**/

static KibaLauncherIconGroup* kiba_launcher_get_group_for_object (KibaLauncher *launcher,
								  GtkWidget    *widget) {
	GList *l;
	
	for (l = launcher->groups; l != NULL; l = l->next) {
		KibaLauncherIconGroup *group = l->data;
		
		if (group->widget == widget)
			return group;
	}
	
	return NULL;
}

static KibaLauncherIconGroup* kiba_launcher_get_group_for_icon (KibaLauncher    *launcher,
								KibaDesktopIcon *icon) {
	GList *l;
	
	for (l = launcher->groups; l != NULL; l = l->next) {
		KibaLauncherIconGroup *group = l->data;
		
		if (strcmp (group->id, icon->entry.group) == 0)
			return group;
	}
	
	return NULL;
}

static void kiba_glist_scroll (GList              **list,
			       GdkScrollDirection   direction) {
	GList *first, *last;
	
	first = *list;
	last = g_list_last (*list);
	
	if (direction == GDK_SCROLL_DOWN) {
		*list = last;
		first->prev = last;
		last->prev->next = NULL;
		last->prev = NULL;
		last->next = first;
	}
	else {
		*list = first->next;
		(*list)->prev = NULL;
		last->next = first;
		first->prev = last;
		first->next = NULL;
	}
}

static void kiba_launcher_ungroup (KibaLauncher *launcher) {
	GList *l;
	
	for (l = launcher->groups; l != NULL; l = l->next) {
		KibaLauncherIconGroup *group = l->data;
		KibaDesktopIcon *group_icon;
		GList *li;
		
		if (KIBA_IS_ICON_VIEW_WIN(group->view_win)) {
			g_signal_handlers_disconnect_by_func (G_OBJECT(group->widget),
				G_CALLBACK(kiba_launcher_icon_view_size_allocate_cb), group);
			g_signal_handlers_disconnect_by_func (G_OBJECT(group->view_win->icon_view),
				G_CALLBACK(kiba_launcher_icon_view_size_allocate_cb), group);
			
			gtk_widget_destroy (group->view_win->win);
			group->view_win = NULL;
		}
		
		for (li = g_list_last(group->icons); li->prev != NULL; li = li->prev) {
			KibaDesktopIcon *icon = li->data;
			
			icon->object = NULL;
			
			group->icons = g_list_remove (group->icons, li->data);
			group->count--;
		}
		
		group_icon = group->icons->data;
		g_free (group->id);
		group->id = g_strdup (group_icon->entry.name);
	}
	
	for (l = launcher->icons; l != NULL; l = l->next) {
		KibaDesktopIcon *icon = l->data;
		
		if (icon->object != NULL)
			continue;
		
		kiba_launcher_create_group_for_icon (launcher, icon);
	}
}

static void kiba_launcher_group (KibaLauncher *launcher) {
	GList *l;
	GList *groups = NULL;
	
	for (l = launcher->groups; l != NULL; l = l->next) {
		KibaLauncherIconGroup *group = l->data;
		KibaDesktopIcon *group_icon;
		GList *li;
		
		group_icon = group->icons->data;
		
		for (li = groups; li != NULL; li = li->next) {
			KibaLauncherIconGroup *check_group = li->data;
			
			if (strcmp (check_group->id, group_icon->entry.group) == 0) {
				group->icons = g_list_remove (group->icons, l->data);
				g_list_free (group->icons);
				kiba_object_destroy (KIBA_OBJECT(group->widget));
				
				g_object_unref (group->launcher);
				
				g_free (group->id);
				g_free (group);
				group = NULL;
				
				kiba_desktop_icon_set_state (group_icon,
					KIBA_DESKTOP_STATE_INACTIVE);
				kiba_desktop_icon_set_object (group_icon,
							check_group->widget);
				
				check_group->icons = g_list_append (check_group->icons, group_icon);
				check_group->count++;
				
				break;
			}
		}
		
		if (!group)
			continue;
		
		g_free (group->id);
		group->id = g_strdup (group_icon->entry.group);
		
		groups = g_list_append (groups, group);
	}
	
	for (l = launcher->groups; l != NULL; l = l->next)
		launcher->groups = g_list_remove (launcher->groups, l->data);
	g_list_free (launcher->groups);
	launcher->groups = groups;
}

static KibaDesktopIcon* kiba_launcher_add_icon_from_uri (KibaLauncher          *launcher,
						         KibaLauncherIconGroup *group,
						         GdkDragAction          action,
						         const char            *launcherdir,
						         const char            *uri) {
	static const char *file_prefix = "file://";
	static const char desktop_entry_group[] = "Desktop Entry";
	
	KibaDesktopIcon *icon;
	FILE *file;
	char *filename = NULL;
	gboolean intern_dnd = FALSE;
	gboolean intern_success = FALSE;
	GList *l;
	
	if (g_strrstr(uri, "/"))
		filename = g_strdup(g_strrstr(uri, "/")+1);
	if (!filename)
		filename = g_strdup(uri);
	
	char path[strlen(launcherdir) + strlen(filename) +1];
	sprintf (path, "%s/%s", launcherdir, filename);
	g_free (filename);
	
	char from_uri[strlen(uri)];
	sprintf (from_uri, "%s", uri);
	
	if (g_str_has_prefix (from_uri, file_prefix)) {
		gint i;
		gint diff;
		gint length;
		
		diff = strlen(file_prefix);
		length = strlen(from_uri);
		
		for (i = 0; i < length; i++) {
			if (i < length - diff)
				from_uri[i] = from_uri[i+diff];
			else {
				from_uri[i] = '\0';
				break;
			}
		}
	}
	
	if (g_file_test(path, G_FILE_TEST_EXISTS)) {
		if (!g_str_has_prefix(path, launcherdir))
			return NULL;
		else
			intern_dnd = TRUE;
	}
	else {
		switch (action) {
			case GDK_ACTION_COPY: {
				char *content;
				
				file = fopen (path, "aw");
				
				if (file == NULL)
					return NULL;
					
				if (!g_file_get_contents (from_uri, &content, NULL, NULL)) {
					fclose (file);
					g_remove (path);
					return NULL;
				}
				
				fprintf (file, content);
				
				g_free (content);
				fclose (file);
				
				break;
			}
			case GDK_ACTION_LINK: {
				link (from_uri, path);
				break;
			}
			case GDK_ACTION_MOVE: {
				g_rename (from_uri, path);
				break;
			}
			default:
				return;
				break;
		}
	}
	
	if (intern_dnd) {
		if (!group)
			return NULL;
		
		for (l = launcher->groups; l != NULL; l = l->next) {
			KibaLauncherIconGroup *check_group = l->data;
			GList *li;
			
			if (check_group == group)
				continue;
			
			for (li = check_group->icons; li != NULL; li = li->next) {
				KibaDesktopIcon *check_icon = KIBA_DESKTOP_ICON(li->data);
				
				if (strcmp (uri, check_icon->entry.uri) == 0) {
					kiba_launcher_remove (launcher, check_icon);
					kiba_desktop_icon_destroy (check_icon);
					intern_success = TRUE;
					break;
				}
			}
		}
	}
	if (intern_dnd && !intern_success)
		return NULL;
	
	if (group) {
		kiba_key_file_write_desktop_entry (path,
			"LauncherGroup", group->id);
	}
	
	icon = kiba_desktop_icon_new (G_TYPE_MODULE(launcher->plugin));
	kiba_desktop_icon_set_file (icon, path);
	kiba_launcher_add (launcher, icon);
	
	kiba_object_switch_surface (KIBA_OBJECT(icon->object));
	
	if (!group)
		group = kiba_launcher_get_group_for_object (launcher, icon->object);
	/* activate icon */
	kiba_desktop_icon_set_state (KIBA_DESKTOP_ICON(group->icons->data),
			KIBA_DESKTOP_STATE_INACTIVE);
	while (group->icons->data != icon)
		kiba_glist_scroll (&group->icons, GDK_SCROLL_DOWN);
	kiba_desktop_icon_set_state (KIBA_DESKTOP_ICON(group->icons->data),
			KIBA_DESKTOP_STATE_ACTIVE);
	
	if (KIBA_IS_ICON_VIEW_WIN(group->view_win)) {
		KibaIconView *icon_view;
		
		icon_view = KIBA_ICON_VIEW(group->view_win->icon_view);
		for (l = gtk_container_get_children (GTK_CONTAINER(icon_view)); l != NULL; l = l->next) {
			GtkWidget *child = l->data;
			KibaDesktopIcon *check_icon;
			
			check_icon = g_object_get_data (G_OBJECT(child), "desktop-icon");
			if (check_icon == icon) {
				kiba_icon_view_select_child (
					KIBA_ICON_VIEW(group->view_win->icon_view),
					child);
				break;
			}
		}
	}
	
	return icon;
}

static gboolean kiba_launcher_drag_receive_cb (KibaLauncher          *launcher,
					       KibaLauncherIconGroup *group,
					       GdkDragContext        *drag_context,
					       const char            *drag_data) {
	KibaDesktopIcon *icon;
	char *string;
	const char *homedir;
	
	homedir = g_get_home_dir ();
	
	char launcher_dir[strlen(homedir)+strlen(LAUNCHER_DIR)];
	sprintf (launcher_dir, "%s" LAUNCHER_DIR, homedir);
	
	char data[strlen(drag_data)];
	sprintf(data, "%s", drag_data);
	g_strstrip (data);
	
	while ((string = g_strrstr (data, "\n")) != NULL) {
		if (!g_str_has_suffix (string, ".desktop")) {
			string[0] = '\0';
			continue;
		}
		
		icon = kiba_launcher_add_icon_from_uri (launcher,
			group,
			drag_context->suggested_action,
			launcher_dir, string+1);
		
		string[0] = '\0';
	}
	
	if (g_str_has_suffix (data, ".desktop")) {
		icon = kiba_launcher_add_icon_from_uri (launcher,
			group,
			drag_context->suggested_action, launcher_dir, data);
	}
	
	return FALSE;
}

static void kiba_launcher_activate_icon (KibaLauncherIconGroup *group,
					 KibaDesktopIcon       *icon,
					 gint                   button) {
	if (icon->entry.exec == NULL)
		return;
		
	kiba_object_launch (KIBA_OBJECT(group->widget), icon->entry.exec);
}

/**
Properties
**/

void kiba_launcher_set_flag (KibaLauncher *launcher,
			     gint          flag) {
	g_return_if_fail (KIBA_IS_LAUNCHER(launcher));
	
	if (GET_FLAG(launcher->flags, flag))
		return;
	
	SET_FLAG(launcher->flags, flag);
	
	switch (flag) {
		case KIBA_LAUNCHER_GROUPING:
			kiba_launcher_group (launcher);
			break;
		default:
			break;
	}
}

void kiba_launcher_unset_flag (KibaLauncher *launcher,
			       gint          flag) {
	g_return_if_fail (KIBA_IS_LAUNCHER(launcher));
	
	if (!GET_FLAG(launcher->flags, flag))
		return;
	
	DEL_FLAG(launcher->flags, flag);
	
	switch (flag) {
		case KIBA_LAUNCHER_GROUPING:
			kiba_launcher_ungroup (launcher);
			break;
		default:
			break;
	}
}

/**
IconView
**/

static void kiba_launcher_group_place_icon_view (KibaLauncherIconGroup *group) {
	KibaWin *win;
	gint w, h, x, y;
	const KibaLayoutData *d;
	
	win = kiba_win_get_default();
	
	if (!KIBA_IS_WIN(win))
		return;
	if (!KIBA_IS_ICON_VIEW_WIN(group->view_win))
		return;
	
	d = &kiba_layout_data[KIBA_DOCK(KIBA_WIN_CHILD(win))->d.position];
	
	struct {
		gint  *x, *y;
		gint  *w, *h;
		gint  alignment;
	} magic[2] = {
		{ &x, &y , &w, &h, d->valignment},
		{ &y, &x , &h, &w, d->halignment}
	};
	w = group->view_win->win->allocation.width;
	h = group->view_win->win->allocation.height;
	x = group->widget->allocation.x + GTK_WIDGET(win)->allocation.x;
	y = group->widget->allocation.y + GTK_WIDGET(win)->allocation.y;
	
	*magic[d->orientation].x -= 
		(*magic[d->orientation].w - group->widget->allocation.width)/2;
	*magic[d->orientation].y -=
		(magic[d->orientation].alignment == 0 ?
			- group->widget->allocation.height : *magic[d->orientation].h);
	
	gdk_window_move (group->view_win->win->window, x, y);
}

static void kiba_launcher_icon_view_size_allocate_cb (GtkWidget     *widget,
						      GtkAllocation *allocation,
						      gpointer       data) {
	KibaLauncherIconGroup *group = (KibaLauncherIconGroup*) data;
	
	kiba_launcher_group_place_icon_view (group);
}

static void kiba_icon_view_child_activated_cb (KibaIconView *icon_view,
					       GtkWidget    *child,
					       gpointer      user_data) {
	KibaLauncherIconGroup *group = user_data;
	
	kiba_launcher_activate_icon (group,
		g_object_get_data (G_OBJECT(child), "desktop-icon"),
		2);
}

static void kiba_icon_view_child_selected_cb (KibaIconView *icon_view,
					      GtkWidget    *child,
					      gpointer      user_data) {
	KibaLauncherIconGroup *group = user_data;
	
	kiba_launcher_select_icon_cb(child, user_data);
}

static void kiba_icon_view_drag_delete_cb (GtkWidget      *widget,
					   GdkDragContext *drag_context,
					   gpointer        user_data) {
	KibaDesktopIcon *icon = user_data;
	
	kiba_launcher_remove (kiba_launcher_get(), icon);
}

static void kiba_icon_view_drag_data_get_cb (GtkWidget        *widget,
					     GdkDragContext   *drag_context,
					     GtkSelectionData *data,
					     guint             info,
					     guint             time,
					     gpointer          user_data) {
	KibaDesktopIcon *icon = user_data;
	
	gtk_selection_data_set(data,
		GDK_SELECTION_TYPE_STRING,
		8,	/* 8 bits per character. */
		(guchar *)icon->entry.uri,
		strlen(icon->entry.uri));
}

static void kiba_icon_view_drag_received_cb (GtkWidget        *widget,
					     GdkDragContext   *drag_context,
					     gint              x,
					     gint              y,
					     GtkSelectionData *selection_data,
					     guint             info,
					     guint             time,
					     gpointer          user_data) {
	KibaLauncherIconGroup *group = user_data;
	
	kiba_launcher_drag_receive_cb (group->launcher,
		group,
		drag_context,
		(char *) selection_data->data);
	
	gtk_drag_finish (drag_context, TRUE, FALSE, time);
}

static GtkWidget *kiba_launcher_create_view_icon (KibaDesktopIcon *icon) {
	static const GtkTargetEntry targets[] = {
		{ "text/uri-list", 0, 0 }
	};
	
	GtkWidget *image;
	GtkWidget *box;
	GtkWidget *label;
	char *markup;
	
	box = gtk_vbox_new(FALSE, 0);
	
	image = kiba_image_new ();
	kiba_image_set_surface (KIBA_IMAGE(image), icon->surface->surface,
			icon->surface->width, icon->surface->height);
	kiba_image_set_size (KIBA_IMAGE(image), 38, 38);
	gtk_widget_show (image);
	gtk_box_pack_start (GTK_BOX (box), image, FALSE, FALSE, 0);
	
	label = gtk_label_new(icon->entry.name);
	gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);
	gtk_label_set_ellipsize (GTK_LABEL(label), PANGO_ELLIPSIZE_END);
	markup = g_markup_printf_escaped ("<span size=\"small\">%s</span>",
			GTK_LABEL (label)->text);
	gtk_label_set_markup (GTK_LABEL (label), markup);
	g_free (markup);
	gtk_widget_show (label);
	gtk_box_pack_start (GTK_BOX (box), label, FALSE, FALSE, 0);
	
	g_object_set_data (G_OBJECT(box), "desktop-icon", icon);
	
	gtk_drag_source_set (image,
		GDK_BUTTON1_MASK | GDK_BUTTON2_MASK,
		targets, G_N_ELEMENTS (targets),
		GDK_ACTION_COPY | GDK_ACTION_LINK | GDK_ACTION_MOVE);
	
	g_signal_connect (GTK_OBJECT(image), "drag-data-get",
		G_CALLBACK(kiba_icon_view_drag_data_get_cb), icon);
	g_signal_connect (GTK_OBJECT(image), "drag-data-delete",
		G_CALLBACK(kiba_icon_view_drag_delete_cb), icon);
	
	return box;
}

static KibaIconViewWin* kiba_launcher_create_icon_view_win (KibaLauncherIconGroup *group) {
	static const GtkTargetEntry targets[] = {
		{ "text/uri-list", 0, 0 }
	};
	
	KibaIconViewWin *view_win;
	GtkWidget *box;
	GList *l;
	
	view_win = kiba_icon_view_win_new(KIBA_ICON_VIEW_BOX_POSITION_RIGHT, 4, 3, FALSE);
	gtk_window_set_type_hint (GTK_WINDOW(view_win->win), GDK_WINDOW_TYPE_HINT_DOCK);
	gtk_window_stick (GTK_WINDOW(view_win->win));
	gtk_window_set_role (GTK_WINDOW(view_win->win), "kiba-launcher");
	for (l = group->icons; l != NULL; l = l->next) {
		KibaDesktopIcon *icon = l->data;
		
		box = kiba_launcher_create_view_icon(icon);
		kiba_icon_view_win_add (view_win, box);
	}
	
	gtk_drag_dest_set (view_win->icon_view,
		GTK_DEST_DEFAULT_ALL,
		targets, G_N_ELEMENTS (targets),
		GDK_ACTION_COPY | GDK_ACTION_LINK | GDK_ACTION_MOVE);
	gtk_drag_dest_add_uri_targets (view_win->icon_view);
	
	g_signal_connect (GTK_OBJECT(view_win->icon_view), "drag-data-received",
		G_CALLBACK(kiba_icon_view_drag_received_cb), group);
	g_signal_connect (GTK_OBJECT(view_win->icon_view), "child-activated",
		G_CALLBACK(kiba_icon_view_child_activated_cb), group);
	g_signal_connect (GTK_OBJECT(view_win->icon_view), "child-selected",
		G_CALLBACK(kiba_icon_view_child_selected_cb), group);
	
	return view_win;
}

static void kiba_launcher_view_win_destroy_cb (GtkWidget *widget,
					       gpointer   user_data) {
	KibaLauncherIconGroup *group = user_data;
	
	group->view_win = NULL;
}

static void kiba_launcher_group_toggle_icon_view (KibaLauncherIconGroup *group) {
	GList *l;
	
	if (KIBA_IS_ICON_VIEW_WIN(group->view_win)) {
		g_signal_handlers_disconnect_by_func (G_OBJECT(group->widget),
			  G_CALLBACK(kiba_launcher_icon_view_size_allocate_cb), group);
		g_signal_handlers_disconnect_by_func (G_OBJECT(group->view_win->icon_view),
			  G_CALLBACK(kiba_launcher_icon_view_size_allocate_cb), group);
		
		gtk_widget_destroy (group->view_win->win);
		group->view_win = NULL;
	}
	else if (GET_FLAG(group->launcher->flags, KIBA_LAUNCHER_GROUPING)) {
		for (l = group->launcher->groups; l != NULL; l = l->next) {
			KibaLauncherIconGroup *check_group = l->data;
			
			if (KIBA_IS_ICON_VIEW_WIN(check_group->view_win)) {
				gtk_widget_destroy (check_group->view_win->win);
				check_group->view_win = NULL;
			}
		}
		group->view_win = kiba_launcher_create_icon_view_win (group);
		gtk_widget_show (group->view_win->win);
		gdk_window_set_role (group->view_win->win->window, "kiba-launcher");
		gdk_window_set_keep_above (group->view_win->win->window, TRUE);
		kiba_launcher_group_place_icon_view (group);
		
		g_signal_connect(G_OBJECT (group->view_win->win), "destroy",
			G_CALLBACK(kiba_launcher_view_win_destroy_cb), group);
		g_signal_connect_after(G_OBJECT (group->widget), "size-allocate",
			G_CALLBACK(kiba_launcher_icon_view_size_allocate_cb), group);
		g_signal_connect_after(G_OBJECT (group->view_win->icon_view), "size-allocate",
			G_CALLBACK(kiba_launcher_icon_view_size_allocate_cb), group);
	}
}

/**
Callbacks
**/

static void update_preview_cb (GtkFileChooser *file_chooser,
			       gpointer        data) {
	GtkWidget *preview;
	GdkPixbuf *pixbuf;
	const char *filename;
	
	preview = GTK_WIDGET(data);
	
	filename = gtk_file_chooser_get_preview_filename(file_chooser);
	
	if (filename != NULL) {
		pixbuf = gdk_pixbuf_new_from_file_at_size(filename, 128, 128, NULL);
		
		if (pixbuf != NULL) {
			gtk_image_set_from_pixbuf(GTK_IMAGE(preview), pixbuf);
			gdk_pixbuf_unref(pixbuf);
		}
		
		gtk_file_chooser_set_preview_widget_active(file_chooser, pixbuf != NULL);
	}
}

static gboolean kiba_launcher_select_icon_cb (GtkWidget *widget,
					      gpointer   data) {
	KibaLauncherIconGroup *group = data;
	KibaIconView *icon_view;
	KibaDesktopIcon *icon;
	GList *l;
	
	icon = g_object_get_data (G_OBJECT(widget), "desktop-icon");
	
	if (group->icons->data == icon)
		return TRUE;
	
	kiba_object_switch_surface (KIBA_OBJECT(icon->object));
	
	kiba_desktop_icon_set_state (KIBA_DESKTOP_ICON(group->icons->data),
			KIBA_DESKTOP_STATE_INACTIVE);
	while (group->icons->data != icon)
		kiba_glist_scroll (&group->icons, GDK_SCROLL_DOWN);
	kiba_desktop_icon_set_state (icon, KIBA_DESKTOP_STATE_ACTIVE);
	
	if (GTK_IS_IMAGE_MENU_ITEM(widget) && group->view_win) {
		icon_view = KIBA_ICON_VIEW(group->view_win->icon_view);
		for (l = gtk_container_get_children (GTK_CONTAINER(icon_view)); l != NULL; l = l->next) {
			GtkWidget *child = l->data;
			KibaDesktopIcon *check_icon;
			
			check_icon = g_object_get_data (G_OBJECT(child), "desktop-icon");
			if (check_icon == icon) {
				kiba_icon_view_select_child (
					KIBA_ICON_VIEW(group->view_win->icon_view),
					child);
				break;
			}
		}
	}
	
	return TRUE;
}

static gboolean kiba_launcher_remove_cb (GtkWidget *widget,
				         gpointer   data) {
	KibaLauncherIconGroup *group = data;
	KibaDesktopIcon *icon;
	const char *homedir;
	
	homedir = g_get_home_dir ();
	
	char launcher_dir[strlen(homedir)+strlen(LAUNCHER_DIR)];
	sprintf (launcher_dir, "%s" LAUNCHER_DIR, homedir);
	
	icon = g_object_get_data (G_OBJECT(widget), "desktop-icon");
	
	if (group->icons->data == data && group->count > 1)
		kiba_object_switch_surface (KIBA_OBJECT(icon->object));
	
	kiba_launcher_remove (group->launcher, icon);
	if (g_file_test(icon->entry.uri, G_FILE_TEST_IS_SYMLINK)) {
		g_unlink (icon->entry.uri);
	}
	else if (g_file_test(icon->entry.uri, G_FILE_TEST_IS_REGULAR)) {
		g_remove (icon->entry.uri);
	}
	if (g_strrstr (icon->entry.icon, launcher_dir)) {
		g_remove (icon->entry.icon);
	}
	kiba_desktop_icon_destroy (icon);
	
	return TRUE;
}

static void desktop_editor_file_saved_cb (KibaDesktopEditor *editor,
					  const char        *file,
					  gpointer           data) {
	KibaDesktopIcon *saved_icon = data;
	KibaLauncherIconGroup *group = NULL;
	KibaLauncher *launcher;
	GList *l;
	
	launcher = kiba_launcher_get();
	
	kiba_launcher_remove (launcher, saved_icon);
	saved_icon->object = NULL;
	kiba_desktop_icon_set_file (saved_icon, file);
	kiba_launcher_add (launcher, saved_icon);
	
	group = kiba_launcher_get_group_for_icon (launcher, saved_icon);
	if (!group)
		return;
	
	kiba_desktop_icon_set_state (KIBA_DESKTOP_ICON(group->icons->data),
			KIBA_DESKTOP_STATE_INACTIVE);
	while (group->icons->data != saved_icon)
		kiba_glist_scroll (&group->icons, GDK_SCROLL_DOWN);
	kiba_desktop_icon_set_state (KIBA_DESKTOP_ICON(group->icons->data),
		KIBA_DESKTOP_STATE_ACTIVE);
}

static void kiba_launcher_edid_cb (GtkWidget      *widget,
				   gpointer        data) {
	KibaDesktopIcon *icon = data;
	KibaDesktopEditor *desktop_editor;
	
	desktop_editor = kiba_desktop_editor_new (icon->entry.uri);
	gtk_widget_show (desktop_editor->win);
	
	g_signal_connect (G_OBJECT(desktop_editor), "file-saved",
		G_CALLBACK(desktop_editor_file_saved_cb), icon);
}

static gboolean kiba_launcher_button_press_cb (GtkWidget      *widget,
					       GdkEventButton *event,
					       gpointer        data) {
	KibaLauncher *launcher = data;
	KibaLauncherIconGroup *group;
	GtkWidget *menu;
	GtkWidget *item, *image;
	GList *l;
	
	if (event->button != 3)
		return FALSE;
	
	menu = gtk_menu_new ();
	
	group = kiba_launcher_get_group_for_object (launcher, widget);
	for (l = group->icons; l != NULL; l = l->next) {
		KibaDesktopIcon *icon = l->data;
		GtkWidget *groupmenu;
		GdkPixbuf *pixbuf = NULL;
		
		kiba_load_icon (icon->entry.icon, 24,  &pixbuf);
		
		if (group->count > 1) {
			groupmenu = gtk_menu_new ();
			
			item = gtk_image_menu_item_new_with_label (icon->entry.name);
			if (pixbuf) {
				image = gtk_image_new_from_pixbuf (pixbuf);
				gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
			}
			
			gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
			gtk_menu_item_set_submenu (GTK_MENU_ITEM (item), GTK_WIDGET(groupmenu));
			gtk_widget_show (item);
		}
		else
			groupmenu = menu;
		
		item = gtk_image_menu_item_new_with_label (_("Edit"));
		image = gtk_image_new_from_stock (GTK_STOCK_EDIT, GTK_ICON_SIZE_DIALOG);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
		gtk_menu_shell_append (GTK_MENU_SHELL (groupmenu), item);
		g_signal_connect (G_OBJECT (item), "activate",
			G_CALLBACK (kiba_launcher_edid_cb), icon);
		gtk_widget_show (item);
		
		kiba_menu_add_separator (groupmenu);
		
		if (group->icons->data != icon) {
			item = gtk_image_menu_item_new_with_label (_("Activate"));
			g_object_set_data (G_OBJECT(item), "desktop-icon", icon);
			image = gtk_image_new_from_icon_name (GTK_STOCK_SELECT_ALL, GTK_ICON_SIZE_MENU);
			gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
			gtk_menu_shell_append (GTK_MENU_SHELL (groupmenu), item);
			g_signal_connect (G_OBJECT (item), "activate",
				G_CALLBACK (kiba_launcher_select_icon_cb), group);
			gtk_widget_show (item);
		}
	
		item = gtk_image_menu_item_new_with_label (_("Launch"));
		image = gtk_image_new_from_icon_name (GTK_STOCK_EXECUTE, GTK_ICON_SIZE_MENU);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
		gtk_menu_shell_append (GTK_MENU_SHELL (groupmenu), item);
		g_signal_connect (G_OBJECT (item), "activate",
			G_CALLBACK (kiba_menu_item_activate_desktop_file), icon->entry.uri);
		gtk_widget_show (item);
		
		item = gtk_image_menu_item_new_with_label (_("Remove"));
		g_object_set_data (G_OBJECT(item), "desktop-icon", icon);
		image = gtk_image_new_from_icon_name (GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
		gtk_menu_shell_append (GTK_MENU_SHELL (groupmenu), item);
		g_signal_connect (G_OBJECT (item), "activate",
			G_CALLBACK (kiba_launcher_remove_cb), group);
		gtk_widget_show (item);
		
		if (pixbuf) {
			g_object_unref (pixbuf);
		}
	}
	
	gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL,
		event->button, event->time);
	
	return TRUE;
}

static gboolean kiba_launcher_button_release_cb (GtkWidget      *widget,
					         GdkEventButton *event,
					         gpointer        data) {
	KibaLauncher *launcher = data;
	KibaDesktopIcon *icon;
	KibaLauncherIconGroup *group;
	
	if (event->button == 3)
		return TRUE;
	
	group = kiba_launcher_get_group_for_object (launcher, widget);
	icon = g_list_nth_data (group->icons, 0);
	
	if (event->button == 2 && GET_FLAG(group->launcher->flags, KIBA_LAUNCHER_GROUPING)) {
		kiba_launcher_group_toggle_icon_view (group);
		return TRUE;
	}
	
	kiba_launcher_activate_icon (group, icon, event->button);
	
	return TRUE;
}

static void kiba_launcher_scroll_cb (GtkWidget      *widget,
				     GdkEventScroll *event,
				     gpointer        data) {
	KibaLauncher *launcher = data;
	KibaLauncherIconGroup *group;
	GList *l;
	
	group = kiba_launcher_get_group_for_object (launcher, widget);
	
	if (g_list_length(group->icons) <= 1)
		return;
	
	kiba_object_switch_surface (KIBA_OBJECT(widget));
	
	kiba_desktop_icon_set_state (KIBA_DESKTOP_ICON(group->icons->data),
			KIBA_DESKTOP_STATE_INACTIVE);
	kiba_glist_scroll (&group->icons, event->direction);
	kiba_desktop_icon_set_state (KIBA_DESKTOP_ICON(group->icons->data),
		KIBA_DESKTOP_STATE_ACTIVE);
	
	if (KIBA_IS_ICON_VIEW_WIN(group->view_win)) {
		kiba_icon_view_scroll_cb (group->view_win->icon_view,
			event, group->view_win);
	}
}

static void kiba_launcher_drag_dest_func (GdkDragContext *drag_context,
				          const char     *drag_data,
				          gpointer        user_data,
				          char          **action_ret,
				          char          **icon_ret) {
	char *string;
	gboolean desktop_file_found;
	
	char data[strlen(drag_data)];
	sprintf(data, "%s", drag_data);
	g_strstrip (data);
	
	desktop_file_found = FALSE;
	while ((string = g_strrstr (data, "\n")) != NULL) {
		if (!g_str_has_suffix (string, ".desktop")) {
			string[0] = '\0';
			continue;
		}
		
		desktop_file_found = TRUE;
		
		string[0] = '\0';
	}
	if (g_str_has_suffix (data, ".desktop"))
		desktop_file_found = TRUE;
	
	if (!desktop_file_found)
		return;
	
	*action_ret = g_strdup(_("Create Launcher"));
	*icon_ret   = g_strdup("launcher-program");
}

static gboolean kiba_launcher_drag_receive_func (GdkDragContext *drag_context,
						 const char     *drag_data,
						 gpointer        user_data) {
	KibaLauncher *launcher = user_data;
	
	return kiba_launcher_drag_receive_cb (launcher, NULL, drag_context, drag_data);
}

static void kiba_object_drag_received_cb (GtkWidget        *widget,
					  GdkDragContext   *drag_context,
					  gint              x,
					  gint              y,
					  GtkSelectionData *selection_data,
					  guint             info,
					  guint             time,
					  gpointer          user_data) {
	KibaLauncher *launcher = user_data;
	KibaLauncherIconGroup *group;
	
	if (GET_FLAG(launcher->flags, KIBA_LAUNCHER_GROUPING))
		group = kiba_launcher_get_group_for_object (launcher, widget);
	else
		group = NULL;
	
	kiba_launcher_drag_receive_cb (launcher,
		group,
		drag_context,
		(char *) selection_data->data);
	
	gtk_drag_finish (drag_context, TRUE, FALSE, time);
}

static void kiba_launcher_create_group_for_icon (KibaLauncher    *launcher,
						 KibaDesktopIcon *icon) {
	KibaLauncherIconGroup *group;
	GtkWidget *widget;
	
	group = g_new0 (KibaLauncherIconGroup, 1);
	if (GET_FLAG(launcher->flags, KIBA_LAUNCHER_GROUPING))
		group->id = g_strdup (icon->entry.group);
	else
		group->id = g_strdup (icon->entry.name);
	group->count = 1;
	group->launcher = g_object_ref (launcher);
	group->icons = NULL;
	group->icons = g_list_append (group->icons, icon);
	group->view_win = NULL;
	
	launcher->groups = g_list_append (launcher->groups, group);
	
	widget = kiba_icon_new(LAUNCHER_PLUGIN_NAME);
	group->widget = widget;
	kiba_object_set_alpha (KIBA_OBJECT(widget), launcher->alpha);
	gtk_widget_show (widget);
	kiba_object_set_requisition_factor (KIBA_OBJECT(widget),
		launcher->size_scale, launcher->size_scale);
	
	g_signal_connect_after (G_OBJECT (widget), "button-press-event",
			G_CALLBACK (kiba_launcher_button_press_cb), launcher);
	g_signal_connect_after (G_OBJECT (widget), "button-release-event",
			G_CALLBACK (kiba_launcher_button_release_cb), launcher);
	g_signal_connect (G_OBJECT (widget), "scroll-event",
			G_CALLBACK (kiba_launcher_scroll_cb), launcher);
	g_signal_connect (G_OBJECT (widget), "drag-data-received",
			G_CALLBACK (kiba_object_drag_received_cb), launcher);
	
	kiba_desktop_icon_set_state (icon, KIBA_DESKTOP_STATE_ACTIVE);
	kiba_desktop_icon_set_object (icon, widget);
}

static void kiba_launcher_icon_file_set_cb (KibaDesktopIcon *icon,
					    const char      *uri,
					    gpointer         data) {
	KibaLauncher *launcher = data;
	KibaLauncherIconGroup *group;
	GList *l;
	
	if (!uri)
		return;
	
	if (GET_FLAG(launcher->flags, KIBA_LAUNCHER_GROUPING)) {
		for (l = launcher->groups; l != NULL; l = l->next) {
			group = l->data;
			
			if (strcmp (group->id, icon->entry.group) == 0) {
				KibaDesktopIcon *object_icon;
				GList *li;
				
				if (g_list_find (group->icons, icon))
					return;
				
				object_icon = g_list_nth_data(group->icons, 0);
				kiba_desktop_icon_set_object (icon, object_icon->object);
				
				group->icons = g_list_append (group->icons, icon);
				group->count++;
				
				for (li = launcher->groups; li != NULL; li = li->next) {
					KibaLauncherIconGroup *check_group = li->data;
					
					if (group == check_group)
						continue;
					
					if (!g_list_find (check_group->icons, icon))
						continue;
					
					check_group->icons = g_list_remove (check_group->icons, icon);
					check_group->count--;
				}
				
				return;
			}
		}
	}
	
	kiba_launcher_create_group_for_icon (launcher, icon);
}

static GList* kiba_launcher_sort_groups (KibaDock     *dock,
					 const char   *group,
					 const GList  *icons,
					 gpointer      data) {
	KibaLauncher *launcher = data;
	GList *ret = NULL;
	const GList *l;
	
	/* sort icons by group (A-Z - a-z)*/
	for (l = icons; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		KibaLauncherIconGroup *group;
		GList *li, *needle = NULL;
		
		group = kiba_launcher_get_group_for_object (launcher, child->widget);
		if (!group) {
			ret = g_list_insert_before (ret, NULL, l->data);
			continue;
		}
		
		for (li = ret; li != NULL; li = li->next) {
			GtkBoxChild *buf_child = li->data;
			KibaLauncherIconGroup *check_group;
			
			check_group = kiba_launcher_get_group_for_object (launcher, buf_child->widget);
			if (!check_group)
				continue;
			
			if (check_group->id[0] > group->id[0]) {
				needle = li;
				break;
			}
			else if (check_group->id[0] == group->id[0]) {
				gint i;
				gint max;
				
				max = MAX(strlen(group->id)+2, strlen(check_group->id)+2);
				for (i = 1; i < max; i++) {
					if (check_group->id[i] > group->id[i]) {
						needle = li;
						break;
					}
					if (check_group->id[i] != group->id[i])
						break;
				}
			}
		}
		
		ret = g_list_insert_before (ret, needle, l->data);
	}
	
	return ret;
}

static void kiba_launcher_icon_app_selected (KibaAppChooser *chooser,
					     const char     *desktop,
					     gpointer        user_data) {
	KibaLauncher *launcher = user_data;
	KibaLauncherIconGroup *group;
	KibaDesktopIcon *icon;
	char *content;
	const char *homedir;
	char *file_name;
	FILE *file;
	
	homedir = g_get_home_dir ();
	
	file_name = g_strrstr (desktop, "/") + 1;
	
	char launcher_dir[strlen(homedir)+strlen(LAUNCHER_DIR)];
	sprintf (launcher_dir, "%s" LAUNCHER_DIR, homedir);
	
	char uri[strlen(launcher_dir) + strlen(file_name) + 1];
	sprintf (uri, "%s/%s", launcher_dir, file_name);
	
	file = fopen (uri, "aw");
	
	if (file == NULL)
		return;
		
	if (!g_file_get_contents (desktop, &content, NULL, NULL)) {
		fclose (file);
		g_remove (uri);
		return;
	}
	
	fprintf (file, content);
	
	g_free (content);
	fclose (file);
	
	icon = kiba_desktop_icon_new (G_TYPE_MODULE(launcher->plugin));
	if (kiba_desktop_icon_set_file (icon, uri))
		kiba_launcher_add (launcher, icon);
	else
		kiba_desktop_icon_destroy (icon);
	
	group = kiba_launcher_get_group_for_object (launcher, icon->object);
	/* activate icon */
	kiba_desktop_icon_set_state (KIBA_DESKTOP_ICON(group->icons->data),
			KIBA_DESKTOP_STATE_INACTIVE);
	while (group->icons->data != icon)
		kiba_glist_scroll (&group->icons, GDK_SCROLL_DOWN);
	kiba_desktop_icon_set_state (KIBA_DESKTOP_ICON(group->icons->data),
			KIBA_DESKTOP_STATE_ACTIVE);
}

static gboolean kiba_launcher_populate_filename_selected (GtkWidget      *button,
							  GdkEventButton *event,
						          gpointer        data) {
	KibaLauncher *launcher = data;
	KibaDesktopIcon *icon;
	KibaDesktopEditor *desktop_editor;
	GtkWidget *entry;
	const char *homedir;
	FILE *launcher_file;
	
	entry = g_object_get_data (G_OBJECT(button), "kiba-entry");
	
	static const char *desktop_suffix = ".desktop";
	
	const char *filename = gtk_entry_get_text (GTK_ENTRY(entry));
	
	if (!filename || strlen(filename) == 0)
		return FALSE;
	
	homedir = g_get_home_dir ();
	
	char launcher_dir[strlen(homedir)+strlen(LAUNCHER_DIR)];
	sprintf (launcher_dir, "%s" LAUNCHER_DIR, homedir);
	
	char *desktop_file = g_strdup_printf ("%s/%s%s", launcher_dir, filename, 
		(g_str_has_suffix (filename, ".desktop") ? "" : desktop_suffix));
	
	if (g_file_test(desktop_file, G_FILE_TEST_EXISTS)) {
		return FALSE;
	}
	
	launcher_file = fopen (desktop_file, "w");
	if (!launcher_file)
		return FALSE;
	
	fprintf (launcher_file,
		"[Desktop Entry]\n"
		"Encoding=UTF-8\n"
		"Version=0.9.2\n"
		"Name=%s\n"
		"Comment=\n"
		"Description=\n"
		"Exec=%s\n"
		"Icon=%s\n"
		"LauncherGroup=%s\n"
		"Class=NotYetLaunched\n",
		filename, filename, filename, filename);
	
	fclose (launcher_file);
	
	icon = kiba_desktop_icon_new (G_TYPE_MODULE(launcher->plugin));
	if (!kiba_desktop_icon_set_file (icon, desktop_file)) {
		kiba_desktop_icon_destroy (icon);
		return FALSE;
	}
	
	kiba_launcher_add (launcher, icon);
	
	desktop_editor = kiba_desktop_editor_new (desktop_file);
	gtk_widget_show (desktop_editor->win);
	
	g_signal_connect (G_OBJECT(desktop_editor), "file-saved",
		G_CALLBACK(desktop_editor_file_saved_cb), icon);
	
	return TRUE;
}

static gboolean kiba_launcher_populate_dialog_destroy (GtkWidget *button,
						       GdkEventButton *event,
						       gpointer   data) {
	GtkWidget *win = data;
	
	gtk_widget_destroy (win);
	
	return TRUE;
}

static void kiba_launcher_populate_dock_create (KibaDock *dock,
					        gpointer  data) {
	GtkWidget *win;
	GtkWidget *box;
	GtkWidget *label;
	GtkWidget *entry;
	GtkWidget *button;
	GtkWidget *button_box;
	GdkPixbuf *pixbuf = NULL;
	
	win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title (GTK_WINDOW(win), _("Launcher creation dialog"));
	gtk_window_set_skip_taskbar_hint (GTK_WINDOW(win), TRUE);
	gtk_window_set_position (GTK_WINDOW(win), GTK_WIN_POS_CENTER);
	if (kiba_load_icon ("kiba-dock", 16, &pixbuf)) {
		gtk_window_set_icon (GTK_WINDOW(win), pixbuf);
		g_object_unref (G_OBJECT(pixbuf));
	}
	
	box = gtk_vbox_new(FALSE, 5); {
		label = gtk_label_new(_("Please enter a launcher name."));
		gtk_box_pack_start (GTK_BOX(box), label, TRUE, TRUE, 0);
		gtk_widget_show (label);
		
		entry = gtk_entry_new ();
		gtk_box_pack_start (GTK_BOX(box), entry, TRUE, TRUE, 0);
		gtk_widget_show (entry);
		
		button_box = gtk_vbox_new(FALSE, 5); {
			button = gtk_button_new_from_stock (GTK_STOCK_OK);
			gtk_box_pack_end (GTK_BOX(button_box), button, FALSE, FALSE, 0);
			g_object_set_data (G_OBJECT(button), "kiba-entry", entry);
			gtk_widget_show (button);
			
			g_signal_connect (G_OBJECT(button), "button-press-event",
				G_CALLBACK(kiba_launcher_populate_filename_selected), data);
			g_signal_connect (G_OBJECT(button), "button-release-event",
				G_CALLBACK(kiba_launcher_populate_dialog_destroy), win);
		}
		gtk_box_pack_start (GTK_BOX(box), button_box, TRUE, TRUE, 0);
		gtk_widget_show (button_box);
	}
	gtk_container_add (GTK_CONTAINER(win), box);
	gtk_widget_show (box);
	gtk_window_present (GTK_WINDOW(win));
}

static void kiba_launcher_populate_dock (KibaDock *dock,
					 gpointer  data) {
	KibaLauncher *launcher = data;
	KibaAppChooser *chooser;
	
	chooser = kiba_app_chooser_new("/usr/share/applications");
	gtk_widget_show (chooser->view_win->win);
	gdk_window_set_keep_above (chooser->view_win->win->window, TRUE);
	
	g_signal_connect (G_OBJECT(chooser), "app-activate",
		G_CALLBACK(kiba_launcher_icon_app_selected), launcher);
}

/**
Public Functions
**/

void kiba_launcher_add (KibaLauncher    *launcher,
			KibaDesktopIcon *icon) {
	KibaLauncherIconGroup *group;
	
	g_return_if_fail(KIBA_IS_LAUNCHER(launcher));
	g_return_if_fail(KIBA_IS_DESKTOP_ICON(icon));
	
	launcher->icons = g_list_append (launcher->icons, icon);
	
	g_signal_connect (G_OBJECT(icon), "file-set",
		G_CALLBACK(kiba_launcher_icon_file_set_cb), launcher);
	kiba_launcher_icon_file_set_cb (icon, icon->entry.uri, launcher);
	
	group = kiba_launcher_get_group_for_object(launcher, icon->object);
	if (!group)
		return;
	
	if (KIBA_IS_ICON_VIEW_WIN(group->view_win)) {
		GtkWidget *box;
		box = kiba_launcher_create_view_icon(icon);
		kiba_icon_view_win_add (group->view_win, box);
	}
	
	g_signal_emit (launcher, launcher->launcher_signals[ICON_ADDED_EVENT], 0, icon);
}

void kiba_launcher_remove (KibaLauncher    *launcher,
			   KibaDesktopIcon *icon) {
	KibaLauncherIconGroup *group;
	KibaIconView *icon_view;
	gboolean first;
	GList *l;
	
	group = kiba_launcher_get_group_for_object(launcher, icon->object);
	g_return_if_fail (group != NULL);
	if (group->count == 0)
		return;
	
	g_signal_handlers_disconnect_by_func (
		     G_OBJECT(icon),
		     G_CALLBACK(kiba_launcher_icon_file_set_cb),
		     launcher);
	
	g_signal_emit (launcher,
		launcher->launcher_signals[ICON_REMOVED_EVENT],
		0, icon);
	
	launcher->icons = g_list_remove (launcher->icons, icon);
	first = (group->icons->data == icon);
	
	group->icons = g_list_remove (group->icons, icon);
	group->count--;
	
	if (KIBA_IS_ICON_VIEW_WIN(group->view_win)) {
		icon_view = KIBA_ICON_VIEW(group->view_win->icon_view);
		
		for (l = gtk_container_get_children(GTK_CONTAINER(icon_view)); l != NULL; l = l->next) {
			GtkWidget *child = l->data;
			KibaDesktopIcon *check_icon;
			
			check_icon = g_object_get_data (G_OBJECT(child), "desktop-icon");
			if (check_icon == icon) {
				if (g_list_length (l) == 1) {
					g_signal_handlers_disconnect_by_func (
					     G_OBJECT(group->widget),
					     G_CALLBACK(kiba_launcher_icon_view_size_allocate_cb),
					     group);
					g_signal_handlers_disconnect_by_func (
					     G_OBJECT(group->view_win->icon_view),
					     G_CALLBACK(kiba_launcher_icon_view_size_allocate_cb),
					     group);
				}
				
				kiba_icon_view_win_remove (group->view_win, child);
				break;
			}
		}
		
		if (!group->count) {
			gtk_widget_destroy (group->view_win->win);
			group->view_win = NULL;
		}
	}
	
	/* the removed object was the last one in the group */
	if (!group->count) {
		kiba_object_destroy (KIBA_OBJECT(icon->object));
		g_list_free (group->icons);
		group->icons = NULL;
		launcher->groups = g_list_remove (launcher->groups, group);
		g_object_unref (group->launcher);
		g_free (group->id);
		group = NULL;
	}
	else if (first) {
		kiba_object_switch_surface (KIBA_OBJECT(icon->object));
		
		if (group) {
			kiba_desktop_icon_set_state (KIBA_DESKTOP_ICON(group->icons->data),
				KIBA_DESKTOP_STATE_ACTIVE);
		}
	}
}

KibaDesktopIcon* kiba_launcher_get_icon_for_widget (KibaLauncher *launcher,
						    GtkWidget    *widget) {
	GList *l;
	
	g_return_val_if_fail (KIBA_IS_LAUNCHER(launcher), NULL);
	g_return_val_if_fail (GTK_IS_WIDGET(widget), NULL);
	
	for (l = launcher->icons; l != NULL; l = l->next) {
		KibaDesktopIcon *icon = l->data;
		
		if (icon->object == widget &&
		    kiba_desktop_icon_get_state(icon) == KIBA_DESKTOP_STATE_ACTIVE)
			return icon;
	}
	
	return NULL;
}

/**
Class Functions
**/

static void kiba_launcher_class_init (KibaLauncherClass *class) {
	kiba_launcher_signals[ICON_ADDED_EVENT] =
		g_signal_new ("icon-added",
				G_OBJECT_CLASS_TYPE (class),
				G_SIGNAL_RUN_LAST,
				G_STRUCT_OFFSET (KibaLauncherClass, icon_added),
				NULL, NULL,
				g_cclosure_marshal_VOID__OBJECT,
				G_TYPE_NONE, 1,
				G_TYPE_OBJECT);
	
	kiba_launcher_signals[ICON_REMOVED_EVENT] =
		g_signal_new ("icon-removed",
				G_OBJECT_CLASS_TYPE (class),
				G_SIGNAL_RUN_LAST,
				G_STRUCT_OFFSET (KibaLauncherClass, icon_removed),
				NULL, NULL,
				g_cclosure_marshal_VOID__OBJECT,
				G_TYPE_NONE, 1,
				G_TYPE_OBJECT);
}

static void kiba_launcher_init (KibaLauncher *launcher) {
	KibaDock *dock;
	
	dock = kiba_dock_get ();
	
	launcher->groups = NULL;
	launcher->icons  = NULL;
	
	launcher->flags = KIBA_LAUNCHER_FLAG_NONE;
	launcher->size_scale = 1.0;
	launcher->alpha = 0.8;
	
	launcher->launcher_signals = kiba_launcher_signals;
	
	kiba_win_add_drag_func (kiba_win_get_default(),
		kiba_launcher_drag_receive_func,
		kiba_launcher_drag_dest_func,
		launcher);
	kiba_dock_set_group_sort_func (dock,
		LAUNCHER_PLUGIN_NAME,
		(KibaGroupSortFunc) kiba_launcher_sort_groups,
		(gpointer) launcher);
	kiba_dock_add_populate_func (dock,
		(KibaDockPopulateFunc) kiba_launcher_populate_dock,
		"launcher-program",
		_("Add Launcher"),
		(gpointer) launcher);
	kiba_dock_add_populate_func (dock,
		(KibaDockPopulateFunc) kiba_launcher_populate_dock_create,
		"launcher-program",
		_("Create Launcher"),
		(gpointer) launcher);
}

KibaLauncher* kiba_launcher_new (GTypeModule *type_module) {
	KibaLauncher *launcher;
	
	kiba_launcher_register_type (type_module);
	launcher = g_object_new (KIBA_TYPE_LAUNCHER, NULL);
	launcher->plugin = KIBA_PLUGIN(type_module);
	
	return launcher;
}

KibaLauncher* kiba_launcher_get (void) {
	KibaPlugin *plugin;
	KibaLauncher *launcher;
 
	plugin = kiba_plugin_get_for_id (LAUNCHER_PLUGIN_NAME);
	if (!plugin)
		return NULL;
	
	launcher = g_object_get_data (G_OBJECT(plugin), LAUNCHER_PLUGIN_NAME);
	if (!launcher)
		return NULL;
	
	if (!kiba_launcher_type_id)
		kiba_launcher_register_type (G_TYPE_MODULE(plugin));
 
	return launcher;
}

void kiba_launcher_destroy (KibaLauncher *launcher) {
	GList *l;
	KibaDock *dock;
	
	dock = kiba_dock_get ();
	
	g_return_if_fail (KIBA_IS_LAUNCHER(launcher));
	
	for (l = launcher->icons; l != NULL; l = launcher->icons) {
		kiba_launcher_remove(launcher, KIBA_DESKTOP_ICON(l->data));
	}
	g_list_free (launcher->icons);
	g_list_free (launcher->groups);
	
	g_object_unref (launcher);
	
	kiba_win_remove_drag_func (kiba_win_get_default(),
			     kiba_launcher_drag_receive_func);
	kiba_dock_remove_group_sort_func (dock,
			LAUNCHER_PLUGIN_NAME);
	kiba_dock_remove_populate_func (dock,
		(KibaDockPopulateFunc) kiba_launcher_populate_dock);
	kiba_dock_remove_populate_func (dock,
		(KibaDockPopulateFunc) kiba_launcher_populate_dock_create);
}
