/*
 *  Copyright (C) 2007  Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <libintl.h>
#include <math.h>
#include <gtk/gtk.h>

#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-object.h>
#include <kiba-fixed-object.h>
#include <kiba-plugin.h>
#include <kiba-prefs.h>
#include <kiba-debug.h>

#define _(string)   gettext (string)
#define PLUGIN_NAME "rotate"

typedef struct _KibaRotate       KibaRotate;
typedef struct _KibaRotateObject KibaRotateObject;

struct _KibaRotateObject {
	KibaRotate  *rotate;
	GtkWidget   *widget;
	
	gboolean     entered;
	gboolean     urgent;
	
	gint         speed;
	gint         timeout_id;
	float       angle;  /* [0.0,2.0] */
	float       scale_factor;
};

typedef enum {
	KIBA_ROTATE_URGENT = 0,
	KIBA_ROTATE_HOVERING,
	KIBA_ROTATE_TRIGGERING
}KibaRotateFlags;

struct _KibaRotate {
	GList  *objects;
	KibaRotateFlags flags;
	
	/* settings */
	gint     speed;
	gint     exec_speed;
};

/**
Helper Functions
**/

static KibaRotateObject* kiba_rotate_get_object_for_widget (KibaRotate  *kiba_rotate,
							    GtkWidget *widget) {
	GList *l;
	
	for (l = kiba_rotate->objects; l != NULL; l = l->next) {
		KibaRotateObject *object = l->data;
		
		if (object->widget == widget)
			return object;
	}
	
	return NULL;
}

static gint rotate_object_timeout (gpointer data) {
	KibaRotateObject *object = data;
	
	object->angle += 0.1*((float) object->speed)/50.0;
	kiba_object_queue_draw (KIBA_OBJECT(object->widget));
	
	if (object->angle >= 2.0) {
		gboolean stop_rotate = FALSE;
		gboolean urgent;
		gboolean hovering;
		
		hovering = (GET_FLAG(object->rotate->flags, KIBA_ROTATE_HOVERING) && object->entered);
		if (!hovering) {
			urgent = (GET_FLAG(object->rotate->flags, KIBA_ROTATE_URGENT) && object->urgent);
			if (!urgent) {
				stop_rotate = TRUE;
			}
		}
		
		object->angle = 0.0;
		
		if (object->speed == object->rotate->exec_speed)
			object->speed = object->rotate->speed;
		
		if (stop_rotate || GET_FLAG(KIBA_OBJECT(object->widget)->state, KIBA_OBJECT_DRAGGING)) {
			object->timeout_id = 0;
			return FALSE;
		}
	}
	
	return TRUE;
}

static void kiba_object_start_rotation (KibaRotateObject *object,
					gint              speed) {
	if (speed < 1)
		return;
	
	if (GET_FLAG(KIBA_OBJECT(object->widget)->state, KIBA_OBJECT_DRAGGING))
		return;
	
	object->speed = speed;
	if (!object->entered)
		object->entered = TRUE;
	
	if (!object->timeout_id)
		object->timeout_id = g_timeout_add (20, rotate_object_timeout, object);
}

/**
Signal Handler
**/

static void kiba_object_scale_func (KibaObject *object,
				    float     *scale_x,
				    float     *scale_y,
				    gpointer    data) {
	KibaRotateObject *rotate_object = data;
	
	*scale_x = rotate_object->scale_factor;
	*scale_y = rotate_object->scale_factor;
}

static void kiba_object_max_scale_func (KibaObject *object,
				        float     *scale_x,
				        float     *scale_y,
				        gpointer    data) {
	KibaRotateObject *rotate_object = data;
	
	*scale_x = 1.0;
	*scale_y = 1.0;
}

static void kiba_rotate_object (KibaObject *object,
				float     *angle,
				gpointer    data) {
	KibaRotateObject *rotate_object = data;
	
	*angle = rotate_object->angle;
}

static void object_urgent_event_cb (KibaObject *object,
				    gboolean    urgent,
				    gpointer    data) {
	KibaRotateObject *rotate = data;
	
	if(urgent) {
		rotate->urgent = TRUE;
		if(GET_FLAG(rotate->rotate->flags, KIBA_ROTATE_URGENT))
			kiba_object_start_rotation(rotate, rotate->rotate->speed);
	} else {
		rotate->urgent = FALSE;
	}
}

static gboolean object_release_event_cb (GtkWidget      *widget,
					 GdkEventButton *event,
					 gpointer        data) {
	KibaRotateObject *object = data;
	
	if (GET_FLAG(object->rotate->flags, KIBA_ROTATE_TRIGGERING))
		kiba_object_start_rotation(object, object->rotate->exec_speed);
	
	return FALSE;
}

static gboolean object_enter_event_cb  (GtkWidget        *widget,
					GdkEventCrossing *event,
					gpointer          data) {
	KibaRotateObject *object = data;
	
	if (GET_FLAG(object->rotate->flags, KIBA_ROTATE_HOVERING))
		kiba_object_start_rotation(object, object->rotate->speed);
	
	return FALSE;
}

static gboolean object_leave_event_cb  (GtkWidget        *widget,
					GdkEventCrossing *event,
					gpointer          data) {
	KibaRotateObject *object = data;
	
	object->entered = FALSE;
	
	return FALSE;
}

static void dock_object_add_cb (GtkContainer *container,
				GtkWidget    *widget,
				gpointer      data) {
	KibaRotate *rotate = data;
	KibaRotateObject *object;
	
	if (KIBA_IS_FIXED_OBJECT(widget))
		return;
	if (kiba_rotate_get_object_for_widget (rotate, widget) != NULL)
		return;
	
	object = g_new0 (KibaRotateObject, 1);
	object->widget = widget;
	object->timeout_id = 0;
	object->speed = 0;
	object->angle = 0.0;
	object->scale_factor = 1.0;
	object->rotate = rotate;
	
	rotate->objects = g_list_append (rotate->objects, object);
	
	kiba_object_add_custom_rotate_function (KIBA_OBJECT(widget),
		kiba_rotate_object, object);
	
	g_signal_connect (G_OBJECT(widget), "enter-notify-event",
		G_CALLBACK(object_enter_event_cb), object);
	g_signal_connect (G_OBJECT(widget), "leave-notify-event",
		G_CALLBACK(object_leave_event_cb), object);
	g_signal_connect (G_OBJECT(widget), "button-release-event",
		G_CALLBACK(object_release_event_cb), object);
	g_signal_connect (G_OBJECT(widget), "urgent-toggled",
		G_CALLBACK(object_urgent_event_cb), object);

}

static void dock_object_remove_cb (GtkContainer *container,
				   GtkWidget    *widget,
				   gpointer      data) {
	KibaRotate *rotate = data;
	KibaRotateObject *object;
	
	object = kiba_rotate_get_object_for_widget (rotate, widget);
	
	if (!object)
		return;
	
	kiba_object_remove_custom_rotate_function_by_data (KIBA_OBJECT(widget), object);
	
	if (object->timeout_id)
		g_source_remove (object->timeout_id);
	
	rotate->objects = g_list_remove (rotate->objects, object);
	
	g_free (object);
}

static void kiba_prefs_hovering_cb (const char *option_id,
				       gpointer    value,
				       gpointer    data) {
	KibaRotate *rotate = data;
	gboolean hovering = GPOINTER_TO_INT(value);
	
	if (hovering) {
		if (!GET_FLAG(rotate->flags, KIBA_ROTATE_HOVERING))
			SET_FLAG(rotate->flags, KIBA_ROTATE_HOVERING);
	}
	else if (GET_FLAG(rotate->flags, KIBA_ROTATE_HOVERING))
		DEL_FLAG(rotate->flags, KIBA_ROTATE_HOVERING);
}

static void kiba_prefs_triggering_cb (const char *option_id,
				       gpointer    value,
				       gpointer    data) {
	KibaRotate *rotate = data;
	gboolean triggering = GPOINTER_TO_INT(value);
	
	if (triggering) {
		if (!GET_FLAG(rotate->flags, KIBA_ROTATE_TRIGGERING))
			SET_FLAG(rotate->flags, KIBA_ROTATE_TRIGGERING);
	}
	else if (GET_FLAG(rotate->flags, KIBA_ROTATE_TRIGGERING))
		DEL_FLAG(rotate->flags, KIBA_ROTATE_TRIGGERING);
}

static void kiba_prefs_urgent_cb (const char *option_id,
				       gpointer    value,
				       gpointer    data) {
	KibaRotate *rotate = data;
	gboolean urgent = GPOINTER_TO_INT(value);
	
	if (urgent) {
		if (!GET_FLAG(rotate->flags, KIBA_ROTATE_URGENT))
			SET_FLAG(rotate->flags, KIBA_ROTATE_URGENT);
	}
	else if (GET_FLAG(rotate->flags, KIBA_ROTATE_URGENT))
		DEL_FLAG(rotate->flags, KIBA_ROTATE_URGENT);
}

static void kiba_prefs_rotate_speed_cb (const char *option_id,
				       gpointer    value,
				       gpointer    data) {
	KibaRotate *rotate = data;
	
	rotate->speed = GPOINTER_TO_INT(value);
}

static void kiba_prefs_rotate_exec_speed_cb (const char *option_id,
				       gpointer    value,
				       gpointer    data) {
	KibaRotate *rotate = data;
	
	rotate->exec_speed = GPOINTER_TO_INT(value);
}

/**
Plugin Functions
**/

static void plugin_unload (KibaPlugin *plugin) {
	KibaRotate *rotate;
	KibaDock *dock;
	GList *l;
	
	dock = kiba_dock_get ();
	rotate = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	
	for (l = rotate->objects; l != NULL; l = rotate->objects) {
		KibaRotateObject *object = l->data;
		
		if (object->timeout_id)
			g_source_remove (object->timeout_id);
		
		g_signal_handlers_disconnect_by_func (G_OBJECT(object->widget),
			G_CALLBACK(object_enter_event_cb), object);
		g_signal_handlers_disconnect_by_func (G_OBJECT(object->widget),
			G_CALLBACK(object_leave_event_cb), object);
		g_signal_handlers_disconnect_by_func (G_OBJECT(object->widget),
			G_CALLBACK(object_release_event_cb), object);
		g_signal_handlers_disconnect_by_func (G_OBJECT(object->widget),
			G_CALLBACK(object_urgent_event_cb), object);
		
		kiba_object_remove_custom_rotate_function_by_data (KIBA_OBJECT(object->widget), object);
		
		g_free (object);
		rotate->objects = g_list_remove (rotate->objects, l->data);
	}
	g_list_free (rotate->objects);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_object_add_cb), rotate);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_object_remove_cb), rotate);
	
	g_free (rotate);
}

static gboolean plugin_load (KibaPlugin *plugin) {
	KibaRotate *kiba_rotate;
	KibaDock *dock;
	GtkContainer *container;
	GList *l;
	
	dock = kiba_dock_get ();
	container = GTK_CONTAINER(dock);
	
	kiba_rotate = g_new0(KibaRotate, 1);
	kiba_rotate->speed = 30;
	kiba_rotate->exec_speed = 50;
	kiba_rotate->objects = NULL;
	
	kiba_prefs_add_callback ("rotate_speed",
		kiba_prefs_rotate_speed_cb, kiba_rotate, TRUE);
	kiba_prefs_add_callback ("rotate_exec_speed",
		kiba_prefs_rotate_exec_speed_cb, kiba_rotate, TRUE);
	kiba_prefs_add_callback ("rotate_hovering",
		kiba_prefs_hovering_cb, kiba_rotate, TRUE);
	kiba_prefs_add_callback ("rotate_triggering",
		kiba_prefs_triggering_cb, kiba_rotate, TRUE);
	kiba_prefs_add_callback ("rotate_urgent",
		kiba_prefs_urgent_cb, kiba_rotate, TRUE);
	
	g_signal_connect_after (G_OBJECT(dock), "add",
		G_CALLBACK(dock_object_add_cb), kiba_rotate);
	g_signal_connect_after (G_OBJECT(dock), "remove",
		G_CALLBACK(dock_object_remove_cb), kiba_rotate);
	for (l = gtk_container_get_children(container); l != NULL; l = l->next) {
		GtkWidget *child = l->data;
		
		dock_object_add_cb (container, child, kiba_rotate);
	}
	
	g_object_set_data (G_OBJECT(plugin),
		PLUGIN_NAME, (gpointer) kiba_rotate);
	
	return TRUE;
}

static PrefMember rotate_options[] = {
    { "rotate_hovering", "FALSE", KIBA_BOOLEAN },
    { "rotate_triggering", "TRUE", KIBA_BOOLEAN },
    { "rotate_urgent", "FALSE", KIBA_BOOLEAN },
    { "rotate_speed", "30", KIBA_INT },
    { "rotate_exec_speed", "40", KIBA_INT },
};

static KibaPluginVtable vtable = {
	PLUGIN_NAME,			/* id/name        */
	"Rotate Plugin for Kiba-Dock",	/* summary        */
	NULL,				/* description    */	

	"Daniel Bessler <danielb@kiba-dock.org>",
					/* author         */
	"http://www.kiba-dock.org",	/* homepage       */

	plugin_load,			/* load            */
	plugin_unload,			/* unload          */
};

KIBA_INIT_PLUGIN (vtable, rotate_options);
