/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <string.h>
#include <gtk/gtk.h>
#include <X11/Xatom.h>

#include <kiba-plugin.h>
#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-object.h>
#include <kiba-screen.h>
#include <kiba-prefs.h>
#include <kiba-debug.h>
#include <kiba-x11.h>

#include "eggtraymanager.h"

#define PLUGIN_NAME "tray"
#define TRAY_ICON_SIZE 28

static void kiba_tray_icon_added (EggTrayManager *manager,
				  GtkWidget      *icon,
				  gpointer        data) {
	GtkWidget *box = data;
	GtkWidget *object;
	GtkWidget *event_box;
	
	object = box->parent;
	
	event_box = gtk_event_box_new ();
	gtk_event_box_set_visible_window (GTK_EVENT_BOX (event_box), FALSE);
	
	gtk_widget_set_size_request (icon, TRAY_ICON_SIZE, TRAY_ICON_SIZE);
	gtk_container_add (GTK_CONTAINER(event_box), icon);
	
	gtk_box_pack_end (GTK_BOX (box), event_box, FALSE, FALSE, 0);
	gtk_widget_show (event_box);
	
	icon->window = object->window;
	g_object_ref (icon->window);
	
	gtk_widget_show (icon);
}

static void kiba_tray_icon_removed (EggTrayManager *manager,
				    GtkWidget      *icon,
				    gpointer        data) {
	GtkWidget *box = data;
	
	gtk_container_remove (GTK_CONTAINER(box), icon->parent);
}

static void kiba_tray_icon_message_sent (EggTrayManager *manager,
					 GtkWidget      *icon,
					 const char     *text,
					 glong           id,
					 glong           timeout,
					 gpointer        data) {
	GtkWidget *object = data;
	
	kiba_object_set_tooltip_text(KIBA_OBJECT(object), text);
}

static void kiba_tray_icon_message_cancelled (EggTrayManager *manager,
					      GtkWidget      *icon,
					      glong           id,
					      gpointer        data) {
	GtkWidget *object = data;
	
	kiba_object_set_tooltip_text(KIBA_OBJECT(object), NULL);
}

static gboolean kiba_tray_expose_event (GtkWidget      *widget,
					GdkEventExpose *expose) {
	cairo_t *cr = NULL;
	GtkWidget *child = NULL;
	
	child = gtk_bin_get_child (GTK_BIN (widget));
	
	/* clear the window for transparent background */
	cr = gdk_cairo_create (expose->window);
		if (!cr) return TRUE;
		
		cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
		cairo_set_source_rgba (cr, 0, 0, 0, 0);
		cairo_paint (cr);
	cairo_destroy (cr);
	
	/* render the child widget */
	if (child)
		gtk_container_propagate_expose (GTK_CONTAINER (widget), child,  expose);
	
	return TRUE;
}

/**
Plugin Functions
**/

static void plugin_unload (KibaPlugin *plugin) {
	EggTrayManager *manager;
	GtkWidget *object;
	GtkWidget *window;
	
	manager = g_object_get_data (G_OBJECT(plugin), "egg-tray-managar");
	g_object_unref (G_OBJECT (manager));
	
	object = g_object_get_data (G_OBJECT(plugin), "dock-object");
	gtk_widget_destroy (object);
	
	window = g_object_get_data (G_OBJECT(plugin), "tray-window");
	gtk_widget_destroy (window);
}

static void kiba_tray_window_size_allocate_cb (GtkWidget     *widget,
					       GtkAllocation *allocation,
					       gpointer       data) {
	KibaObject *object = data;
	float scale_w, scale_h;
	
//TODO
// 	scale_w = (float) allocation->width/(*object->size);
// 	scale_h = (float) allocation->height/(*object->size);
// 	
// 	kiba_object_set_requisition_factor (object, scale_w, scale_h);
}

static void kiba_tray_object_size_allocate_cb (GtkWidget     *widget,
					       GtkAllocation *allocation,
					       gpointer       data) {
	GtkWidget *win = data;
	GtkWidget *parent;
	gint x = 0, y = 0;
	
	for (parent = widget; parent->parent != NULL; parent = parent->parent);
	
	if (KIBA_IS_WIN(parent)) {
		x += parent->allocation.x;
		y += parent->allocation.y;
	}
	
	gtk_window_move (GTK_WINDOW(win),
		x + allocation->x,
		y + allocation->y);
}

static gboolean plugin_load (KibaPlugin *plugin) {
	GtkWidget *object;
	GtkWidget *box;
	EggTrayManager *manager;
	
	if (egg_tray_manager_check_running(KIBA_GDK_SCREEN)) {
		DEBUG_ERROR ("another systray already running\n");
		return FALSE;
	}
	
	manager = egg_tray_manager_new (G_TYPE_MODULE(plugin));
	if (!manager) {
		DEBUG_ERROR ("Failed to create System tray Manager Instance\n");
		return FALSE;
	}
	
	if (!egg_tray_manager_manage_screen (manager, KIBA_GDK_SCREEN)) {
		DEBUG_ERROR ("System tray didn't get the system tray manager selection\n");
		g_object_unref (G_OBJECT (manager));
		return FALSE;
	}
	
	object = kiba_object_new (PLUGIN_NAME);
	KIBA_OBJECT(object)->win_state = KIBA_OBJECT_WINDOW_INVISIBLE;
	SET_FLAG (KIBA_OBJECT(object)->props, KIBA_OBJECT_VISIBLE);
	SET_FLAG (KIBA_OBJECT(object)->props, KIBA_OBJECT_MOVEABLE);
	gtk_widget_show (object);
	
	GtkWidget *window;
	
	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_decorated (GTK_WINDOW(window), FALSE);
	gtk_window_set_resizable (GTK_WINDOW(window), FALSE);
	gtk_window_set_skip_taskbar_hint (GTK_WINDOW(window), TRUE);
	gtk_window_set_type_hint (GTK_WINDOW(window), GDK_WINDOW_TYPE_HINT_DOCK);
	
	box = gtk_hbox_new(FALSE, 0);
	gtk_container_set_border_width (GTK_CONTAINER(box), 3);
	gtk_container_add (GTK_CONTAINER(window), box);
	gtk_widget_show (box);
	
	gtk_window_present (GTK_WINDOW(window));
	gdk_window_set_keep_above (window->window, TRUE);
	
// 	g_signal_connect (G_OBJECT(object), "expose-event",
// 			G_CALLBACK (kiba_tray_expose_event), NULL);
	g_signal_connect (G_OBJECT(window), "size-allocate",
			G_CALLBACK (kiba_tray_window_size_allocate_cb), object);
	g_signal_connect (G_OBJECT(object), "size-allocate",
			G_CALLBACK (kiba_tray_object_size_allocate_cb), window);
	g_signal_connect (G_OBJECT(manager), "tray_icon_added",
			G_CALLBACK (kiba_tray_icon_added), box);
	g_signal_connect (G_OBJECT(manager), "tray_icon_removed",
			G_CALLBACK (kiba_tray_icon_removed), box);
	g_signal_connect (G_OBJECT(manager), "message_sent",
			G_CALLBACK (kiba_tray_icon_message_sent), object);
	g_signal_connect (G_OBJECT(manager), "message_cancelled",
			G_CALLBACK (kiba_tray_icon_message_cancelled), object);
	
	g_object_set_data (G_OBJECT(plugin),
		"egg-tray-managar", (gpointer) manager);
	g_object_set_data (G_OBJECT(plugin),
		"dock-object", (gpointer) object);
	g_object_set_data (G_OBJECT(plugin),
		"tray-window", (gpointer) window);
	
	return TRUE;
}

static PrefMember tray_options[] = {
};

static KibaPluginVtable vtable = {
	PLUGIN_NAME,			/* id/name        */
	"Tray Plugin for Kiba-Dock",	/* summary        */
	NULL,				/* description    */	

	"Daniel Bessler <danielb@kiba-dock.org>",
					/* author         */
	"http://www.kiba-dock.org",	/* homepage       */

	plugin_load,			/* load            */
	plugin_unload,			/* unload          */
};

KIBA_INIT_PLUGIN (vtable, tray_options);
