/*
 *  Copyright (C) 2007  Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "../config.h"
#include <libintl.h>
#include <stdlib.h>
#include <gtk/gtk.h>

#include <kiba-dock.h>
#include <kiba-icon.h>
#include <kiba-plugin.h>
#include <kiba-prefs.h>
#include <kiba-debug.h>
#include <kiba-dialog.h>
#include <kiba-macros.h>
#include <kiba-info-win.h>

#include "alsa_mixer.h"

#define _(String)            dgettext (GETTEXT_PACKAGE, String)
#define PLUGIN_NAME          "volume"
#define VOLUME_FALLBACK_ICON "volume"

#define MUTE_LEVEL  -1
#define ZERO_LEVEL   0
#define MIN_LEVEL    33
#define MEDIUM_LEVEL 66
#define MAX_LEVEL    100

typedef struct _KibaVolume KibaVolume;

mixer_ops_t *alsa_mixer;

typedef enum {
	KIBA_VOLUME_MUTED = 0,
	KIBA_VOLUME_ZERO,
	KIBA_VOLUME_MIN,
	KIBA_VOLUME_MEDIUM,
	KIBA_VOLUME_MAX
}KibaVolumeState;

/* keep the same order as KibaVolumeState enum above */
static gint volume_level_data[] = {
	MUTE_LEVEL,    /**< KIBA_VOLUME_MUTED.   */
	ZERO_LEVEL,    /**< KIBA_VOLUME_ZERO.    */
	MIN_LEVEL,     /**< KIBA_VOLUME_MIN.     */
	MEDIUM_LEVEL,  /**< KIBA_VOLUME_MEDIUM.  */
	MAX_LEVEL      /**< KIBA_VOLUME_MAX.     */
};

/* keep the same order as KibaVolumeState enum above */
static const char* volume_icon_data[2][5] = {
	{ "volume-mute", "volume-zero", "volume-min",
	  "volume-medium",  "volume-max" },
	{ "stock_volume-mute", "stock_volume-0", "stock_volume-min",
	  "stock_volume-med",  "stock_volume-max" },
};

struct _KibaVolume {
	GtkWidget       *icon;
	KibaVolumeState  state;
	
	mixer_t         *mixer;       /* pointer to opened mixer device */
	unsigned int     device;
	
	int              pleft;       /* volume left (used when muted) */
	int              pright;      /* volume left (used when muted) */
	
	GtkWidget       *scrollbar;
	GtkRange        *scale;
	GtkButton       *plus;
	GtkButton       *minus;
	gboolean         pop;
	gint             direction;
	gint             adjust_timeout;
	
	int              timeout_id;
	
	/* settings */
	char            *mixer_id;
	char            *device_id;
};

static void kiba_volume_pop_dock (KibaVolume *volume);

static void kiba_volume_popup_dock (KibaVolume *volume);

static GtkWidget* kiba_volume_dock_new (KibaVolume    *volume,
					GtkOrientation orientation);

static void volume_device_cb (const char *option_id,
			      gpointer    value,
			      gpointer    data);

/**
Helper Functions
**/

static void show_no_mixer_dialog (void) {
	static gboolean no_mixer_dialog_shown = FALSE;
	char *message;
	
	if (no_mixer_dialog_shown)
		return;
	no_mixer_dialog_shown = TRUE;
	
	kiba_dialog_error (_("No Mixer found"), "%s\n\n%s",
			_("The volume control did not find any elements and/or "
			"devices to control."),
			_("You can remove the volume control from kiba by "
			"disabling the volume plugin in kiba-settings."));
}

static void kiba_mixer_update (KibaVolume *volume) {
	mixer_idz_t *idz,*t;
	
	idz = mixer_get_id_list();
	for (t = idz; t != NULL; t = t->next) {
		if (volume->mixer_id != NULL &&
		    strcmp (volume->mixer_id, t->id) == 0) {
			break;
		}
		else if (t->next == NULL) {
			if (volume->mixer_id != NULL)
				g_free (volume->mixer_id);
			volume->mixer_id = g_strdup (t->id);
			break;
		}
	}
	mixer_free_idz(idz);
	
	if (volume->mixer_id == NULL) {
		show_no_mixer_dialog ();
		volume->mixer = NULL;
		return;
	}
	
	if (volume->mixer) {
		mixer_close (volume->mixer);
	}
	volume->mixer = mixer_open (volume->mixer_id);
	
	DEBUG_INFO("Using Mixer Device %s\n", volume->mixer_id);
}

static void kiba_volume_update_icon (KibaVolume *volume) {
	gint i;
	
	for (i = 0; i < G_N_ELEMENTS(volume_icon_data); i++) {
		if (kiba_icon_set_file (KIBA_ICON(volume->icon),
			volume_icon_data[i][volume->state]))
			break;
	}
}

static void kiba_volume_update_flag (KibaVolume *volume,
				     int         value) {
	gint level,
	     i;
	
	/* initiall level must be below the lowest level in volume_level_data */
	level = MUTE_LEVEL - 1;
	
	/* get the current level */
	for (i = 0; level < value; i++)
		level = volume_level_data[i];
	i--;
	
	if (i != volume->state) {
		volume->state = i;
		kiba_volume_update_icon (volume);
	}
}

static gboolean volume_dock_button_press_timeout (gpointer data) {
	KibaVolume    *volume = data;
	GtkAdjustment *adj;
	gfloat         volume_value;
	gboolean       res = TRUE;
	
	if (!volume->adjust_timeout)
		return FALSE;
	
	adj = gtk_range_get_adjustment (volume->scale);
	volume_value = gtk_range_get_value (volume->scale);
	volume_value += volume->direction * adj->step_increment;
	
	if (volume_value <= adj->lower) {
		volume_value = adj->lower;
		res = FALSE;
	} else if (volume_value >= adj->upper) {
		volume_value = adj->upper;
		res = FALSE;
	}
	gtk_range_set_value (volume->scale, volume_value);
	
	if (!res)
		volume->adjust_timeout = 0;
	
	return res;
}

/**
check if volume changed in an interval
**/
static gboolean kiba_volume_time_handler (KibaVolume *volume) {
	int right_volume = 0, left_volume  = 0;
	char *volume_text  = NULL;
	
	if (!volume->mixer)
		return FALSE;
	
	mixer_get_device_volume (volume->mixer, volume->device, &left_volume, &right_volume);
	if (volume->state != KIBA_VOLUME_MUTED || left_volume + right_volume > 0)
		kiba_volume_update_flag (volume, (left_volume + right_volume)/2);
	
	if (volume->state == KIBA_VOLUME_MUTED)
		volume_text = g_strdup (_("Mixer muted"));
	else
		volume_text = g_strdup_printf ("%s: %i%%",
			_("Volume at"), left_volume > right_volume ? left_volume : right_volume);
	
	kiba_object_set_title (KIBA_OBJECT(volume->icon), volume_text);
	gtk_range_set_value (volume->scale, (left_volume + right_volume) / 2.0);
	
	g_free (volume_text);
	
	return TRUE;
}

/**
Signal Handler
**/

static void kiba_volume_switch_mixer (GtkWidget *widget,
				       gpointer   data) {
	KibaVolume *volume = data;
	
	const char* mixer_id = gtk_widget_get_name (widget);
	
	if (volume->mixer_id) {
		if (strcmp (mixer_id, volume->mixer_id) == 0)
			return;
		g_free (volume->mixer_id);
	}
	
	volume->mixer_id = g_strdup (mixer_id);
	kiba_mixer_update (volume);
}

static void kiba_volume_switch_device (GtkWidget *widget,
				       gpointer   data) {
	KibaVolume *volume = data;
	
	const char* device_id = gtk_widget_get_name (widget);
	
	if (volume->device_id) {
		if (strcmp (device_id, volume->device_id) == 0)
			return;
	}
	
	volume_device_cb ("volume_device_name", (char*) device_id, volume);
}

static void kiba_volume_mute_mixer (GtkWidget *widget,
				    gpointer   data) {
	KibaVolume *volume = data;
	
	mixer_set_device_volume (volume->mixer, volume->device, 0, 0);
	kiba_volume_update_flag (volume, -1);
}

static void  kiba_volume_unmute_mixer (GtkWidget *widget,
				       gpointer   data) {
	KibaVolume *volume = data;
	
	kiba_volume_update_flag (volume, (volume->pleft + volume->pright) / 2);
	mixer_set_device_volume (volume->mixer, volume->device, volume->pleft, volume->pright);
}

static gboolean cb_button_press (GtkWidget      *widget,
				 GdkEventButton *button,
				 gpointer        data) {
	KibaVolume *volume = data;
	
	volume->direction = (GTK_BUTTON (widget) == volume->plus) ? 1 : -1;
	if (volume->adjust_timeout)
		g_source_remove (volume->adjust_timeout);
	volume->adjust_timeout = g_timeout_add (100, volume_dock_button_press_timeout, data);
	volume_dock_button_press_timeout (data);
	
	return TRUE;
}

static gboolean cb_button_release (GtkWidget      *widget,
				   GdkEventButton *button,
				   gpointer        data) {
	KibaVolume *volume = data;
	
	if (volume->adjust_timeout) {
		g_source_remove (volume->adjust_timeout);
		volume->adjust_timeout = 0;
	}
	
	return TRUE;
}

static void scrollbar_volume_changed_cb (GtkAdjustment *adj,
				         gpointer       data) {
	KibaVolume *volume = data;
	gfloat value, nright, nleft;
	gint pright, pleft;
	char *title;
	
	if (!volume->mixer || volume->state == KIBA_VOLUME_MUTED)
		return;
	
	value = gtk_adjustment_get_value (adj);
	
	mixer_get_device_volume (volume->mixer, volume->device, &pleft, &pright);
	
	nright = (pright - pleft)/2.0 + value;
	nleft = (pleft - pright)/2.0 + value;
	
	if (nright > 100.0) {
		nleft += nright - 100.0;
		nright = 100.0;
	}
	else if (nleft > 100.0) {
		nright += nleft - 100.0;
		nleft = 100.0;
	}
	else if (nright < 0.0) {
		nleft += nright;
		nright = 0.0;
	}
	else if (nleft < 0.0) {
		nright += nleft;
		nleft = 0.0;
	}
	
	if (volume->pleft == nleft && volume->pright == nright)
		return;
	
	mixer_set_device_volume (volume->mixer, volume->device, nleft, nright);
	volume->pleft = nleft;
	volume->pright = nright;
	
	title = g_strdup_printf ("%s: %i%%", _("Volume at"),
					pleft > pright ? pleft : pright);
	
	kiba_object_set_title (KIBA_OBJECT(volume->icon), title);
	g_free (title);
}

static void icon_theme_changed_cb (GtkIconTheme *icon_theme,
				   gpointer      user_data) {
	KibaVolume *volume = user_data;
	
	kiba_volume_update_icon (volume);
}

/**
Volume Dock
**/

static GtkWidget* kiba_volume_dock_new (KibaVolume    *volume,
					GtkOrientation orientation) {
	GtkWidget *table, *button, *scale;
	GtkWidget *dock;
	struct {
		gint w, h;
		gint x[3], y[3];
		GtkWidget * (* sfunc) (GtkAdjustment *adj);
		gint sw, sh;
	} magic[2] = {
		{ 3, 1, { 0, 1, 2 }, { 0, 0, 0 }, gtk_hscale_new, 100, -1 }, /**< GTK_ORIENTATION_HORIZONTAL. */
		{ 1, 3, { 0, 0, 0 }, { 0, 1, 2 }, gtk_vscale_new, -1, 100 }  /**< GTK_ORIENTATION_VERTICAL. */
	};
	
	dock = kiba_info_win_new ();
	gtk_window_set_role (GTK_WINDOW(dock), "kiba-volume");
	gtk_window_set_accept_focus (GTK_WINDOW(dock), TRUE);
	gtk_window_set_keep_above (GTK_WINDOW(dock), TRUE);
	gtk_window_set_decorated (GTK_WINDOW(dock), FALSE);
	gtk_window_set_resizable (GTK_WINDOW(dock), FALSE);
	gtk_window_stick (GTK_WINDOW(dock));
	gtk_window_set_skip_taskbar_hint (GTK_WINDOW(dock), TRUE);
	
	table = gtk_table_new (magic[orientation].w,
				magic[orientation].h, FALSE);
	
	button = gtk_button_new_with_label (_("+"));
	volume->plus = GTK_BUTTON (button);
	gtk_button_set_relief (volume->plus, GTK_RELIEF_NONE);
	gtk_table_attach_defaults (GTK_TABLE (table), button,
				magic[orientation].x[0],
				magic[orientation].x[0] + 1,
				magic[orientation].y[0],
				magic[orientation].y[0] + 1);
	g_signal_connect (button, "button-press-event",
			G_CALLBACK (cb_button_press), volume);
	g_signal_connect (button, "button-release-event",
			G_CALLBACK (cb_button_release), volume);
	gtk_widget_show (button);
	
	scale = magic[orientation].sfunc (NULL);
	volume->scale = GTK_RANGE (scale);
	gtk_widget_set_size_request (scale,
				magic[orientation].sw,
				magic[orientation].sh);
	gtk_scale_set_draw_value (GTK_SCALE (scale), FALSE);
	gtk_range_set_inverted (volume->scale, TRUE);
	gtk_table_attach_defaults (GTK_TABLE (table), scale,
				magic[orientation].x[1],
				magic[orientation].x[1] + 1,
				magic[orientation].y[1],
				magic[orientation].y[1] + 1);
	gtk_widget_show (scale);
	
	button = gtk_button_new_with_label (_("-"));
	volume->minus = GTK_BUTTON (button);
	gtk_button_set_relief (volume->minus, GTK_RELIEF_NONE);
	gtk_table_attach_defaults (GTK_TABLE (table), button,
				magic[orientation].x[2],
				magic[orientation].x[2] + 1,
				magic[orientation].y[2],
				magic[orientation].y[2] + 1);
	g_signal_connect (button, "button-press-event",
			G_CALLBACK (cb_button_press), volume);
	g_signal_connect (button, "button-release-event",
			G_CALLBACK (cb_button_release), volume);
	gtk_widget_show (button);
	
	gtk_container_add (GTK_CONTAINER (dock), table);
	gtk_widget_show (table);
	
	return GTK_WIDGET (dock);
}

static void kiba_volume_get_dock_position (KibaVolume *volume,
					   gint       *x,
					   gint       *y) {
	GtkWidget *dock;
	gint width, height;
	const KibaLayoutData *d;
	
	dock = GTK_WIDGET(volume->icon)->parent;
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	gtk_window_get_size (GTK_WINDOW (volume->scrollbar), &width, &height);
	*x = volume->icon->allocation.x + dock->parent->allocation.x;
	*y = volume->icon->allocation.y + dock->parent->allocation.y;
	
	/* switch coordinates ;) */
	struct {
		gint *x, *y;
		gint  w,  h;
		gint alignment;
	} layout_magic[2] = {
		{ x, y, width,  height, d->valignment },  /**< GTK_ORIENTATION_HORIZONTAL. */
		{ y, x, height, width,  d->halignment }   /**< GTK_ORIENTATION_VERTICAL. */
	};
	
	*layout_magic[d->orientation].x -=
		(layout_magic[d->orientation].w - volume->icon->allocation.width) / 2.0;
	*layout_magic[d->orientation].y -=
		layout_magic[d->orientation].alignment == 0 ?
			-volume->icon->allocation.width : layout_magic[d->orientation].h;
}

static void kiba_volume_place_dock (KibaVolume *volume) {
	gint x, y;
	
	kiba_volume_get_dock_position (volume, &x, &y);
	gtk_window_move (GTK_WINDOW (volume->scrollbar), x, y);
}

static void kiba_volume_popup_dock (KibaVolume *volume) {
	gtk_window_present (GTK_WINDOW (volume->scrollbar));
	
	/* reposition */
	kiba_volume_place_dock (volume);
	
	/* grab input */
	gtk_widget_grab_focus (volume->scrollbar);
	gtk_widget_grab_focus (GTK_WIDGET(volume->scale));
	
	/* keep state */
	volume->pop = TRUE;
}

static void kiba_volume_popdown_dock (KibaVolume *volume) {
	if (!volume->pop)
		return;
	
	/* hide */
	gtk_widget_hide (volume->scrollbar);
	
	/* keep state */
	volume->pop = FALSE;
}

static void kiba_volume_pop_dock (KibaVolume *volume) {
	if (volume->pop)
		kiba_volume_popdown_dock (volume);
	else
		kiba_volume_popup_dock (volume);
}

/**
KibaObject Callbacks
**/

static gboolean kiba_object_button_press_cb (GtkWidget      *widget,
					     GdkEventButton *event,
					     gpointer        data) {
	KibaPlugin *plugin = data;
	KibaVolume *volume;
	GtkWidget *menu;
	GtkWidget *item, *image;
	GList *device_ids = NULL;
	char *device_id = NULL;
	mixer_idz_t  *idz,*t;
	gint i;
	
	volume = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	
	menu = gtk_menu_new ();
	
	idz = mixer_get_id_list();
	for (t = idz; t != NULL && device_id == NULL; t = t->next)
		device_ids = g_list_append (device_ids, g_strdup (t->id));
	mixer_free_idz(idz);
	
	if (g_list_length (device_ids) > 1) {
		GtkMenu *switch_menu;
		GList *l;
		
		switch_menu  = (GtkMenu *) gtk_menu_new();
		
		item = gtk_image_menu_item_new_with_label (_("Switch to another Mixer Device"));
		image = gtk_image_new_from_icon_name (GTK_STOCK_REFRESH, GTK_ICON_SIZE_MENU);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
		gtk_menu_item_set_submenu (GTK_MENU_ITEM (item), GTK_WIDGET(switch_menu));
		gtk_widget_show (item);
		
		for (l = device_ids; l != NULL; l = l->next) {
			char *dev_id = l->data;
			
			if (volume->mixer_id && strcmp (dev_id, volume->mixer_id) == 0)
				continue;
			
			item = gtk_image_menu_item_new_with_label (dev_id);
			image = gtk_image_new_from_icon_name ("volume-zero", GTK_ICON_SIZE_MENU);
			gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
			gtk_widget_set_name (item, dev_id);
			gtk_menu_shell_append (GTK_MENU_SHELL (switch_menu), item);
			g_signal_connect (G_OBJECT (item),
				"activate", G_CALLBACK (kiba_volume_switch_mixer), volume);
			gtk_widget_show (item);
	
			g_free (dev_id);
		}
	}
	
	if (mixer_get_nr_devices(volume->mixer) > 1) {
		GtkMenu *switch_menu;
		
		switch_menu  = (GtkMenu *) gtk_menu_new();
		
		item = gtk_image_menu_item_new_with_label (_("Mixer device"));
		image = gtk_image_new_from_icon_name (GTK_STOCK_REFRESH, GTK_ICON_SIZE_MENU);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
		gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
		gtk_menu_item_set_submenu (GTK_MENU_ITEM (item), GTK_WIDGET(switch_menu));
		gtk_widget_show (item);
		
		for(i = 0; i < mixer_get_nr_devices(volume->mixer); i++) {
			char *name;
			
			if (mixer_get_device_fullscale(volume->mixer, i) == 1) {
				/*Switch not supported yet */
				continue;
			}
			if (i == volume->device)
				continue;
			
			name = mixer_get_device_real_name(volume->mixer,i);
			item = gtk_image_menu_item_new_with_label (name);
			gtk_widget_set_name(item, name);
			gtk_menu_shell_append (GTK_MENU_SHELL (switch_menu), item);
			g_signal_connect (G_OBJECT (item),
				"activate", G_CALLBACK (kiba_volume_switch_device), volume);
			gtk_widget_show (item);
		}
	}
	
	if (volume->state == KIBA_VOLUME_MUTED) {
		item = gtk_image_menu_item_new_with_label (_("Unmute Mixer"));
		image = gtk_image_new_from_icon_name (volume_icon_data[0][KIBA_VOLUME_ZERO],
				GTK_ICON_SIZE_MENU);
		g_signal_connect (G_OBJECT (item),
			"activate", G_CALLBACK (kiba_volume_unmute_mixer), volume);
	}
	else {
		item = gtk_image_menu_item_new_with_label (_("Mute Mixer"));
		image = gtk_image_new_from_icon_name (volume_icon_data[0][KIBA_VOLUME_MUTED],
				GTK_ICON_SIZE_MENU);
		g_signal_connect (G_OBJECT (item),
			"activate", G_CALLBACK (kiba_volume_mute_mixer), volume);
	}
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
	gtk_widget_show (item);
	
	if (device_ids)
		g_list_free (device_ids);
	
	gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL,
		event->button, event->time);
	
	return TRUE;
}

static gboolean kiba_object_button_release_cb (GtkWidget      *widget,
					       GdkEventButton *event,
					       gpointer        data) {
	KibaVolume *volume = data;
	
	if (event->button == 2) {
		if (volume->state == KIBA_VOLUME_MUTED)
			kiba_volume_unmute_mixer (NULL, volume);
		else
			kiba_volume_mute_mixer (NULL, volume);
	}
	else if (event->button == 1)
		kiba_volume_pop_dock (volume);
	
	return FALSE;
}

static void kiba_object_scroll_cb (GtkWidget      *widget,
				   GdkEventScroll *event,
				   gpointer        data) {
	KibaVolume *volume = data;
	int right_volume = 0, left_volume = 0, new_left, new_right;
	char *title;
	const int direction[] = {
		-1, 1
	};
	
	if (!volume->mixer || volume->state == KIBA_VOLUME_MUTED)
		return;
	
	mixer_get_device_volume (volume->mixer, volume->device, &left_volume, &right_volume);
	new_left  = left_volume + direction[event->direction == GDK_SCROLL_UP] * 5;
	new_right = right_volume + direction[event->direction == GDK_SCROLL_UP] * 5;
	new_left  = MAX(0, MIN(100, new_left));
	new_right = MAX(0, MIN(100, new_right));
	mixer_set_device_volume (volume->mixer, volume->device, new_left, new_right);
	volume->pleft  = new_left;
	volume->pright = new_right;
	
	title= g_strdup_printf ("%s: %i%%",
		_("Volume at"), new_left > new_right ? new_left : new_right);
	kiba_object_set_title (KIBA_OBJECT(volume->icon), title);
	g_free (title);
	gtk_range_set_value (volume->scale, (new_left + new_right) / 2.0);
}

static void kiba_object_size_allocate_cb (GtkWidget     *widget,
					  GtkAllocation *allocation,
					  gpointer       user_data) {
	KibaVolume *volume = user_data;
	
	if (!volume->pop)
		return;
	
	kiba_volume_place_dock (volume);
}

/**
KibaPrefs Callbacks
**/

static void dock_object_size_cb (const char *option_id,
			         gpointer    value,
			         gpointer    data) {
	KibaVolume *volume = data;
	
	kiba_volume_update_icon (volume);
}

static void object_size_factor_cb (const char *option_id,
			           gpointer    value,
			           gpointer    data) {
	KibaVolume *volume = data;
	float factor;
	
	factor = g_ascii_strtod((char*) value,NULL);
	kiba_object_set_requisition_factor (KIBA_OBJECT(volume->icon), factor, factor);
	kiba_object_resize (KIBA_OBJECT(volume->icon));
	KIBA_OBJECT(volume->icon)->surface_data->width  = volume->icon->allocation.width;
	KIBA_OBJECT(volume->icon)->surface_data->height = volume->icon->allocation.height;
	kiba_volume_update_icon (volume);
	gtk_widget_queue_resize (volume->icon->parent);
}

static void object_alpha_cb (const char *option_id,
			     gpointer    value,
			     gpointer    data) {
	KibaVolume *volume = data;
	float alpha;
	
	alpha = g_ascii_strtod((char*) value,NULL);
	
	kiba_object_set_alpha (KIBA_OBJECT(volume->icon), alpha);
}

static void volume_device_cb (const char *option_id,
			      gpointer    value,
			      gpointer    data) {
	KibaVolume *volume = data;
	int right_volume = 0, left_volume = 0;
	gint i;
	
	if (volume->device_id)
		g_free (volume->device_id);
	volume->device_id = g_strdup ((char*) value);
	volume->device = 0;
	
	if (volume->device_id) {
		for(i = 0; i < mixer_get_nr_devices(volume->mixer); i++) {
			if (mixer_get_device_fullscale(volume->mixer, i) == 1) {
				/*Switch not supported yet */
				continue;
			}
			
			if (strcmp (volume->device_id, mixer_get_device_real_name(volume->mixer,i)) == 0) {
				volume->device = i;
				break;
			}
		}
	}
	/* using the first valid device */
	if (volume->device == 0) {
		for(i = 0; i < mixer_get_nr_devices(volume->mixer); i++) {
			if (mixer_get_device_fullscale(volume->mixer, i) == 1) {
				/*Switch not supported yet */
				continue;
			}
			volume->device = i;
			if (!volume->device_id) {
				volume->device_id = mixer_get_device_real_name(volume->mixer,i);
			}
			break;
		}
	}
	
	mixer_get_device_volume (volume->mixer, volume->device, &left_volume, &right_volume);
	volume->pleft  = left_volume;
	volume->pright = right_volume;
	kiba_volume_update_flag (volume, (left_volume + right_volume)/2);
}

static void volume_mixer_cb (const char *option_id,
			     gpointer    value,
			     gpointer    data) {
	KibaVolume *volume = data;
	mixer_idz_t *idz,*t;
	gboolean validMixer = FALSE;
	
	if (volume->mixer_id)
		g_free (volume->mixer_id);
	
	idz = mixer_get_id_list();
	if (value != NULL) {
		for (t = idz; t != NULL; t = t->next) {
			if (strcmp ((char*) value, t->id) == 0) {
				validMixer = TRUE;
				break;
			}
		}
	}
	
	if (validMixer)
		volume->mixer_id = g_strdup ((char*) value);
	else
		volume->mixer_id = g_strdup ((char*) idz->id);
	
	mixer_free_idz(idz);
	
	kiba_mixer_update(volume);
}

/**
Plugin Functions
**/

static void plugin_unload (KibaPlugin *plugin) {
	KibaVolume  *volume;
	
	volume = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	
	kiba_prefs_remove_callback_by_data (volume);
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(gtk_icon_theme_get_default()),
		G_CALLBACK(icon_theme_changed_cb), volume);
	g_signal_handlers_disconnect_by_func (G_OBJECT(volume->icon),
		G_CALLBACK(kiba_object_size_allocate_cb), volume);
	
	g_source_remove (volume->timeout_id);
	
	kiba_object_destroy (KIBA_OBJECT(volume->icon));
	
	if (GTK_IS_WIDGET(volume->scrollbar))
		gtk_widget_destroy (volume->scrollbar);
	
	if (volume->mixer_id) {
		kiba_prefs_write_to_config_file ("volume_mixer_id", "char", volume->mixer_id);
		g_free (volume->mixer_id);
	}
	if (volume->device_id) {
		kiba_prefs_write_to_config_file ("volume_device_name", "char", volume->device_id);
		g_free (volume->device_id);
	}
	
	if (volume->mixer)
		mixer_close (volume->mixer);
	
	g_free (volume);
}

static gboolean plugin_load (KibaPlugin *plugin) {
	KibaVolume *volume;
	KibaDock *dock;
	GtkObject *adj;
	int right_volume = 0, left_volume = 0;
	mixer_idz_t *idz;
	const KibaLayoutData *d;
	
#ifdef ENABLE_NLS
	bindtextdomain (GETTEXT_PACKAGE, DATADIR"/locale");
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");	
#endif
	
	dock = kiba_dock_get ();
	
	init_mixer ();
	
	idz = mixer_get_id_list();
	if (idz == NULL) {
		DEBUG_ERROR("No valid Mixer found!\n");
		return FALSE;
	}
	mixer_free_idz(idz);

	volume = g_new0 (KibaVolume, 1);
	volume->state = KIBA_VOLUME_ZERO;
	volume->mixer_id = NULL;
	volume->device_id = NULL;
	volume->device = 1;
	
	/* add icon to the dock */
	volume->icon = kiba_icon_new (PLUGIN_NAME);
	gtk_widget_show (volume->icon);
	
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	/* initial Mixer */
	kiba_prefs_add_callback ("volume_mixer_id",
		volume_mixer_cb, volume, TRUE);
	kiba_prefs_add_callback ("volume_device_name",
		volume_device_cb, volume, TRUE);
	
	/* load icon */
	mixer_get_device_volume (volume->mixer, volume->device, &left_volume, &right_volume);
	volume->pleft = left_volume;
	volume->pright = right_volume;
	kiba_volume_update_flag (volume, (left_volume + right_volume)/2);
	
	adj = gtk_adjustment_new (50, 0, 100, 4, 10, 0);
	gtk_adjustment_set_value (GTK_ADJUSTMENT (adj), (float) (left_volume + right_volume) / 2.0);
	volume->scrollbar = kiba_volume_dock_new (volume, -(d->orientation-1));
	gtk_range_set_adjustment (volume->scale, GTK_ADJUSTMENT (adj));
	g_signal_connect (adj, "value-changed", G_CALLBACK (scrollbar_volume_changed_cb), volume);
	
	volume->timeout_id = g_timeout_add (SECOND_INTERVAL,
		(GSourceFunc) kiba_volume_time_handler, (gpointer) volume);
	
	g_object_set_data (G_OBJECT(plugin),
		PLUGIN_NAME, (gpointer) volume);
	
	/* add object signal handler */
	g_signal_connect (G_OBJECT (volume->icon), "button-press-event",
			G_CALLBACK (kiba_object_button_press_cb), plugin);
	g_signal_connect (G_OBJECT (volume->icon), "button-release-event",
			G_CALLBACK (kiba_object_button_release_cb), volume);
	g_signal_connect (G_OBJECT (volume->icon), "scroll-event",
			G_CALLBACK (kiba_object_scroll_cb), volume);
	g_signal_connect (G_OBJECT (volume->icon), "size-allocate",
			G_CALLBACK (kiba_object_size_allocate_cb), volume);
	
	g_signal_connect (G_OBJECT (gtk_icon_theme_get_default()), "changed",
			G_CALLBACK (icon_theme_changed_cb), volume);
	
	kiba_prefs_add_callback ("dock_object_size",
		dock_object_size_cb, volume, FALSE);
	kiba_prefs_add_callback ("volume_size_factor",
		object_size_factor_cb, volume, TRUE);
	kiba_prefs_add_callback ("volume_alpha",
		object_alpha_cb, volume, TRUE);
	
	return TRUE;
}

static PrefMember volume_options[] = {
	{ "volume_mixer_id", "NULL", KIBA_STRING },
	{ "volume_device_name", "PCM", KIBA_STRING },
	{ "volume_size_factor", "1.0", KIBA_DOUBLE },
	{ "volume_alpha", "1.0", KIBA_DOUBLE },
};

static KibaPluginVtable vtable = {
	PLUGIN_NAME,			/* id/name        */
	"Volume Plugin for Kiba-Dock",	/* summary        */
	NULL,				/* description    */	

	"Daniel Bessler <danielb@kiba-dock.org>",
					/* author         */
	"http://www.kiba-dock.org",	/* homepage       */

	plugin_load,			/* load            */
	plugin_unload,			/* unload          */
};

KIBA_INIT_PLUGIN (vtable, volume_options);
