/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
 *
 *  Note   : based on taskabr plugin
*/

#include <string.h>
#include <stdlib.h>
#include <libintl.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-icon.h>
#include <kiba-plugin.h>
#include <kiba-screen.h>
#include <kiba-window.h>
#include <kiba-prefs.h>
#include <kiba-x11.h>
#include <kiba-debug.h>
#include <kiba-macros.h>
#include <kiba-inline-pixbufs.h>

#define _(String) dgettext (GETTEXT_PACKAGE, String)

#define PLUGIN_NAME "winlist"

typedef struct _KibaWinlist     KibaWinlist;
typedef struct _KibaWinlistIcon KibaWinlistIcon;

struct _KibaWinlistIcon {
	KibaWinlist   *winlist;
	GtkWidget     *object;
	KibaWindow    *window;
	
	CairoSurfaceData  *surface_data;
	gchar             *title;
	gint               winlist_member;
};

struct _KibaWinlist {
	GList    *winlist_icons; /* list of all KibaWinlistIcon's */
	
	GSList   *skip_apps;
	gint      size;
	gint      dock_object_size;
	float    size_scale;
};

static void kiba_winlist_object_new (KibaWinlistIcon *icon,
				     KibaPlugin      *plugin);

static KibaWinlistIcon* kiba_winlist_icon_new (KibaPlugin  *plugin,
					       KibaWindow  *window);

/**
Helper Functions
**/

static KibaWinlistIcon* kiba_winlist_get_active_icon (KibaWinlist *winlist) {
	GList *l;
	
	for (l = winlist->winlist_icons; l != NULL; l = l->next) {
		KibaWinlistIcon *icon = l->data;
	
		if (icon->winlist_member == 1)
			return icon;
	}
	
	return NULL;
}

static KibaWinlistIcon* kiba_winlist_get_icon_for_window (KibaWinlist *winlist,
							  void        *window) {
	GList *l;
	
	for (l = winlist->winlist_icons; l != NULL; l = l->next) {
		KibaWinlistIcon *icon = l->data;
		
		if (icon->window == window)
			return icon;
	}
	
	return FALSE;
}

static gint sort_winlist_icons (gconstpointer a,
				gconstpointer b) {
	const KibaWinlistIcon *aicon = a;
	const KibaWinlistIcon *bicon = b;
	
	return aicon->winlist_member - bicon->winlist_member;
}

static void kiba_winlist_scroll (KibaWinlist *winlist,
				 GtkWidget   *object,
				 gint         direction,
				 gboolean     focus) {
	KibaWinlistIcon *old_active,
		        *new_active = NULL;
	GList *l;
	
	old_active = kiba_winlist_get_active_icon (winlist);
	
	if (g_list_length (winlist->winlist_icons) == 1) {
		kiba_focus_window (NULL, old_active->window->gdk_window);
		return;
	}
	
	/* reorder winlist member */
	for (l = winlist->winlist_icons; l != NULL; l = l->next) {
		KibaWinlistIcon *icon = l->data;
	
		if (direction == 1) {
			if (icon->winlist_member != 1)
				icon->winlist_member--;
			else
				icon->winlist_member = g_list_length (winlist->winlist_icons);
		}
		else if (direction == -1) {
			if (icon->winlist_member != g_list_length (winlist->winlist_icons))
				icon->winlist_member++;
			else
				icon->winlist_member = 1;
		}
	}
	
	winlist->winlist_icons = g_list_sort (winlist->winlist_icons,
			(GCompareFunc)sort_winlist_icons);
	
	new_active = kiba_winlist_get_active_icon (winlist);
	if (!new_active)
		return;
	
	if (new_active == old_active) {
		kiba_focus_window (NULL, old_active->window->gdk_window);
		return;
	}
	
	old_active->object = NULL;
	new_active->object = object;
	
	kiba_object_set_title (KIBA_OBJECT(new_active->object), new_active->title);
	KIBA_OBJECT(new_active->object)->surface_data = new_active->surface_data;
	kiba_object_queue_draw (KIBA_OBJECT(new_active->object));
	
	if (!focus)
		return;
	
	gboolean window_on_active_desktop;
	int x,y, wx, wy;
	
	kiba_get_avtive_viewport (&x, &y);
	kiba_get_win_viewport_number (gdk_x11_drawable_get_xid (new_active->window->gdk_window), &wx, &wy);
	
	window_on_active_desktop = x == wx && y == wy &&
		kiba_get_active_workspace_number (gdk_display) ==
		kiba_get_win_workspace_number (gdk_display,
			gdk_x11_drawable_get_xid (new_active->window->gdk_window));
	
	int win_state;
		
	win_state = kiba_get_winstate (gdk_display, new_active->window->gdk_window);
	
	if (win_state == IconicState) {
		kiba_deiconify_window (NULL, new_active->window->gdk_window);
		kiba_focus_window (NULL, new_active->window->gdk_window);
	}
	else if (win_state == NormalState)
		kiba_focus_window (NULL, new_active->window->gdk_window);
}

static void kiba_winlist_icon_draw_surface (cairo_t          *cr,
					    cairo_surface_t  *surface,
					    gint              width,
					    gint              height,
					    void             *data) {
	KibaWinlistIcon *icon = data;
	GdkPixbuf *pixbuf = NULL;
	
	pixbuf = kiba_get_icon (gdk_x11_drawable_get_xid (icon->window->gdk_window),
				icon->winlist->size);
	if (!pixbuf) {
		pixbuf = gdk_pixbuf_new_from_inline (-1, app_fallback_icon, FALSE, NULL);
		pixbuf = gdk_pixbuf_scale_simple (pixbuf,
				icon->winlist->size, icon->winlist->size,
				GDK_INTERP_BILINEAR);
	}
	
	gdk_cairo_set_source_pixbuf (cr, pixbuf, 0, 0);
	cairo_paint (cr);
	
	g_object_unref (pixbuf);
}

static void kiba_winlist_icon_update_surface (KibaWinlistIcon *icon) {
	KibaWin *default_win;
	cairo_t *cr;
	
	icon->surface_data->width  = icon->winlist->size;
	icon->surface_data->height = icon->winlist->size;
	
	default_win = kiba_win_get_default();
	cr = kiba_win_cairo_create (default_win);
	cairo_update_surface (cr, icon->surface_data,
			kiba_winlist_icon_draw_surface, icon);
	cairo_destroy (cr);
	
	if (icon->object)
		kiba_object_queue_draw (KIBA_OBJECT(icon->object));
}

static gboolean kiba_winlist_check_type (KibaWinlist *winlist,
					 KibaWindow  *window) {
	char *state = kiba_get_wintype (gdk_display, window->gdk_window);
	
	if (strcmp (state, "_NET_WM_WINDOW_TYPE_NORMAL") == 0) {
		g_free (state);
		return TRUE;
	}
	
	g_free (state);
	
	return FALSE;
}

static gboolean kiba_winlist_skip_hints (KibaWinlist *winlist,
					 KibaWindow  *window) {
	GSList *l;
	XTextProperty text_property;
	Atom cmd_atom;
	char *state;
	
	/* skip apps in exclude list */
	cmd_atom = XInternAtom (gdk_display, "WM_CLASS", TRUE);
	if (XGetTextProperty (gdk_display, window->xwindow, &text_property, cmd_atom)) {
		for (l = winlist->skip_apps; l != NULL; l = l->next) {
			char *skip_app = l->data;
			
			if (g_strrstr (text_property.value, skip_app)) {
				XFree (text_property.value);
				return TRUE;
			}
		}
	}

	if (text_property.value)
		XFree (text_property.value);
	
	return kiba_window_state_hint (window->gdk_window, "_NET_WM_STATE_SKIP_TASKBAR");
}

/**
KibaScreen Callbacks
**/

static void kiba_winlist_add_window (KibaWinlist *winlist,
                                     KibaWindow  *window,
                                     KibaPlugin  *plugin) {
	KibaWinlistIcon *icon;
	
	if (kiba_winlist_skip_hints (winlist, window) || !kiba_winlist_check_type (winlist, window))
		return;
	
	icon = kiba_winlist_icon_new (plugin, window);
	if (icon == NULL)
		return;
	
	if (icon->winlist_member == 1 && !icon->object) {
		kiba_winlist_object_new (icon, plugin);
	
		kiba_object_set_icon_geometry (KIBA_OBJECT(icon->object),
				window);
	}
}

static void window_opened_callback (KibaScreen  *screen,
				    KibaWindow  *window,
				    KibaPlugin  *plugin) {
	KibaWinlist *winlist;
	KibaWinlistIcon *icon;
	
	winlist = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	kiba_winlist_add_window (winlist, window, plugin);
}

static void kiba_winlist_remove_window (KibaWinlist *winlist,
                                        KibaWindow  *window) {
	KibaWinlistIcon *icon;
	
	icon = kiba_winlist_get_icon_for_window (winlist, window);
	if (!icon)
		return;
	
	/* remove object when there is just one object in winlist */
	if (icon->object && g_list_length (winlist->winlist_icons) == 1)
		gtk_widget_destroy (icon->object);
	else if (icon->winlist_member == 1 && g_list_length (winlist->winlist_icons) > 1) {
		kiba_winlist_scroll (winlist, icon->object, 1, FALSE);
		if (icon->surface_data->surface != NULL)
			cairo_surface_destroy (icon->surface_data->surface);
		g_free (icon->surface_data);
	}
	
	winlist->winlist_icons = g_list_remove (winlist->winlist_icons, icon);
	
	
	g_free (icon);
	icon = NULL;
}

static void window_closed_callback (KibaScreen  *screen,
				    KibaWindow  *window,
				    KibaPlugin  *plugin) {
	KibaWinlist *winlist;
	KibaWinlistIcon *icon;
	
	winlist = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	kiba_winlist_remove_window (winlist, window);
}

static void withdrawn_toggled_callback (KibaScreen  *screen,
                                        KibaWindow  *window,
                                        KibaPlugin  *plugin) {
	KibaWinlist *winlist;

	winlist = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);

	if (window->state & (1 << KIBA_WINDOW_WITHDRAWN)) {
		kiba_winlist_remove_window (winlist, window);
	} else {
		kiba_winlist_add_window (winlist, window, plugin);
	}
}

static void active_window_changed_callback (KibaScreen  *screen,
					    KibaWindow  *active_window,
					    KibaWinlist *winlist) {
	KibaWinlistIcon *icon;
	GList *l;
	
	icon = kiba_winlist_get_icon_for_window (winlist, active_window);
	if (icon == NULL || icon->winlist_member == 1)
		return;
	
	for (l = winlist->winlist_icons; l != NULL; l = l->next) {
		KibaWinlistIcon *buf_icon = l->data;
		
		if (buf_icon->winlist_member == 1) {
			icon->object = buf_icon->object;
			
			kiba_object_set_title (KIBA_OBJECT(icon->object), icon->title);
			KIBA_OBJECT(icon->object)->surface_data =  icon->surface_data;
			kiba_object_queue_draw (KIBA_OBJECT(icon->object));
		}
		if (buf_icon->winlist_member < icon->winlist_member)
			buf_icon->winlist_member++;
	}
	icon->winlist_member = 1;
	
	winlist->winlist_icons = g_list_sort (winlist->winlist_icons,
			(GCompareFunc)sort_winlist_icons);
}

/**
KibaWindow Callbacks
**/

static void icon_changed_callback (KibaWindow      *window,
				   KibaWinlistIcon *icon) {
	kiba_winlist_icon_update_surface (icon);
}

static void state_changed_callback (KibaWindow      *window,
				    gint             state,
				    KibaWinlistIcon *icon) {	
	if (icon->object) {
		if (GET_FLAG(state, KIBA_WINDOW_URGENT) &&
		   !kiba_object_get_urgent(KIBA_OBJECT(icon->object)))
			kiba_object_set_urgent(KIBA_OBJECT(icon->object), TRUE);
		else if (!GET_FLAG(state, KIBA_WINDOW_URGENT) &&
		         kiba_object_get_urgent(KIBA_OBJECT(icon->object)))
			kiba_object_set_urgent(KIBA_OBJECT(icon->object), FALSE);
	}
}

static void name_changed_callback (KibaWindow      *window,
				   const gchar     *name,
				   KibaWinlistIcon *icon) {
	char  *object_name = g_strdup (name);
	
	while (TRUE) {
		GList *li;
		int    i = 1;
		
		for (li = icon->winlist->winlist_icons; li != NULL; li = li->next) {
			KibaWinlistIcon *check_icon = li->data;
		
			if (strcmp (check_icon->title, object_name) == 0 && check_icon != icon)
				i++;
		}
		
		if (i > 1) {
			int   old_value = 0;
		
			if (g_str_has_suffix (object_name, ")") &&
			    g_strrstr (object_name, " (") != NULL) {
				char *buf, *value;
				
				buf   = g_strrstr (object_name, "(") + 1;
				value = g_strndup (buf, strlen(buf) - 1);
				old_value = atoi (value) - 1;
				g_free (value);
			}
		
			g_free (object_name);
			object_name = NULL;
			object_name = g_strdup_printf ("%s (%i)", name, i + old_value);
		}
		else
			break;
	}
	
	if (icon->title) {
		g_free (icon->title);
		icon->title = NULL;
	}
	icon->title = g_strdup (object_name);
	
	if (icon->object) {
		kiba_object_set_title (KIBA_OBJECT(icon->object), object_name);
	}
	
	g_free (object_name);
}

/**
KibaPrefs Callbacks
**/

static void dock_object_size_cb (const char *option_id,
			         gpointer    value,
			         gpointer    data) {
	KibaWinlist *winlist = data;
	KibaWinlistIcon *icon;
	GList *l;
	
	winlist->dock_object_size = GPOINTER_TO_INT(value);
	winlist->size = winlist->dock_object_size * winlist->size_scale;
	
	for (l = winlist->winlist_icons; l != NULL; l = l->next) {
		icon = l->data;
	
		kiba_winlist_icon_update_surface (icon);
	}
}

static void object_size_factor_cb (const char *option_id,
			           gpointer    value,
			           gpointer    data) {
	KibaWinlist *winlist = data;
	KibaWinlistIcon *icon;
	GList *l;
	
	winlist->size_scale = g_ascii_strtod((char*) value, NULL);
	winlist->size = winlist->dock_object_size * winlist->size_scale;
	icon = kiba_winlist_get_active_icon (winlist);
	
	if (!icon)
		return;
	
	if (icon) {
		kiba_object_set_requisition_factor (KIBA_OBJECT(icon->object),
			winlist->size_scale, winlist->size_scale);
		kiba_object_resize (KIBA_OBJECT(icon->object));
	}
	
	for (l = winlist->winlist_icons; l != NULL; l = l->next) {
		icon = l->data;
	
		kiba_winlist_icon_update_surface (icon);
	}
	gtk_widget_queue_resize (icon->object->parent);
}

static void object_alpha_cb (const char *option_id,
			     gpointer    value,
			     gpointer    data) {
	KibaWinlist *winlist = data;
	KibaWinlistIcon *icon;
	float alpha;
	
	icon = kiba_winlist_get_active_icon (winlist);
	if (!icon)
		return;
	
	alpha = g_ascii_strtod((char*) value, NULL);
	
	kiba_object_set_alpha (KIBA_OBJECT(icon->object), alpha);
}

static void skip_app_cb (const char *option_id,
			 gpointer    value,
			 gpointer    data) {
	KibaWinlist *winlist = data;
	
	winlist->skip_apps = (GSList*) value;
}

/**
KibaObject functions
**/

static gboolean kiba_object_button_press_cb (GtkWidget      *widget,
					     GdkEventButton *event,
					     gpointer        data) {
	KibaPlugin *plugin = data;
	KibaWinlist *winlist;
	GtkWidget *menu;
	GtkWidget *item, *image;
	GList *l;
	
	winlist = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	
	menu = gtk_menu_new ();
	
	for (l = winlist->winlist_icons; l != NULL; l = l->next) {
		KibaWinlistIcon *icon = l->data;
	
		if (icon->window->gdk_window != NULL) {
			GdkPixbuf *pixbuf;
			GtkMenu   *winmenu;
			char      *text;
			
			text = kiba_win_name (gdk_display,
					gdk_x11_drawable_get_xid (icon->window->gdk_window));
			winmenu = GTK_MENU (gtk_menu_new());
			
			item = gtk_image_menu_item_new_with_label (text);
			g_free (text);
			pixbuf = cairo_get_pixbuf_from_surface (icon->surface_data->surface,
					icon->surface_data->width, icon->surface_data->height);
			pixbuf = gdk_pixbuf_scale_simple (pixbuf,
					24, 24, GDK_INTERP_NEAREST);
			image = gtk_image_new_from_pixbuf (pixbuf);
			g_object_unref (pixbuf);
			gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
			gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
			gtk_menu_item_set_submenu (GTK_MENU_ITEM (item), GTK_WIDGET(winmenu));
			gtk_widget_show (item);
			
			kiba_load_app_menu (icon->window->gdk_window, winmenu);
			
			kiba_object_set_icon_geometry (KIBA_OBJECT(widget),
				icon->window);
		}
	}
	
	gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL,
		event->button, event->time);
	
	return TRUE;
}

static gboolean kiba_object_button_release_cb (GtkWidget      *widget,
					       GdkEventButton *event,
					       gpointer        data) {
	KibaWinlist *winlist = data;
	KibaWinlistIcon *icon;
	Window focus_window;
	
	icon = kiba_winlist_get_active_icon (winlist);
	g_return_if_fail (icon != NULL);
	focus_window = kiba_get_active_window (gdk_display);
	
	switch (kiba_get_winstate (gdk_display, icon->window->gdk_window)) {
		case IconicState:  /* window iconified */
			kiba_object_set_icon_geometry (KIBA_OBJECT(icon->object),
				icon->window);
			kiba_deiconify_window (NULL, icon->window->gdk_window);
			break;
		case NormalState: /* window uniconified */
			/* window focussed */
			if (gdk_x11_drawable_get_xid (icon->window->gdk_window) == focus_window) {
				kiba_object_set_icon_geometry (KIBA_OBJECT(icon->object),
					icon->window);
				kiba_iconify_window (NULL, icon->window->gdk_window);
			}
			/* window unfocussed */
			else {
				kiba_focus_window (NULL, icon->window->gdk_window);
			}
		default: /* window unmapped */
			break;
	}
	
	return FALSE;
}

static void kiba_object_scroll_cb (GtkWidget      *widget,
				   GdkEventScroll *event,
				   gpointer        data) {
	KibaWinlist *winlist = data;
	const int direction[] = {
		-1, 1
	};
	
	kiba_winlist_scroll (winlist, widget,
		direction[event->direction == GDK_SCROLL_UP], TRUE);
}

/**
Winlist Init Functions
**/

static void kiba_winlist_object_new (KibaWinlistIcon *icon,
				     KibaPlugin      *plugin) {
	icon->object = kiba_icon_new (PLUGIN_NAME);
	gtk_widget_show (icon->object);
	
	kiba_object_set_requisition_factor (KIBA_OBJECT(icon->object),
		icon->winlist->size_scale, icon->winlist->size_scale);
	kiba_object_resize (KIBA_OBJECT(icon->object));
	
	KIBA_OBJECT(icon->object)->surface_data = icon->surface_data;
	kiba_object_set_title (KIBA_OBJECT(icon->object), icon->title);
	kiba_object_set_active (icon->object, TRUE);
	
	kiba_object_set_alpha (KIBA_OBJECT(icon->object),
		kiba_prefs_get_float ("winlist_alpha"));
	
	/* add object signal handler */
	g_signal_connect (G_OBJECT (icon->object), "button-press-event",
			G_CALLBACK (kiba_object_button_press_cb), plugin);
	g_signal_connect (G_OBJECT (icon->object), "button-release-event",
			G_CALLBACK (kiba_object_button_release_cb), icon->winlist);
	g_signal_connect (G_OBJECT (icon->object), "scroll-event",
			G_CALLBACK (kiba_object_scroll_cb), icon->winlist);
}

static KibaWinlistIcon* kiba_winlist_icon_new (KibaPlugin *plugin,
					       KibaWindow *window) {
	KibaWinlistIcon *icon;
	KibaWinlist *winlist;
	
	g_return_if_fail (KIBA_IS_WINDOW(window));
	
	winlist = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	
	if (!window->name || kiba_winlist_skip_hints (winlist, window) ||
            !kiba_winlist_check_type (winlist, window))
		return NULL;
	
	icon = g_new0 (KibaWinlistIcon, 1);
	icon->winlist = winlist;
	icon->window = window;
	icon->winlist_member = g_list_length (winlist->winlist_icons) + 1;
	icon->object = NULL;
	
	icon->surface_data = g_new0 (CairoSurfaceData, 1);
	icon->surface_data->surface = NULL;
	icon->surface_data->width  = winlist->size;
	icon->surface_data->height = winlist->size;
	
	/* intiall window stuff */
	name_changed_callback (icon->window, icon->window->name, icon);
	state_changed_callback (icon->window, icon->window->state, icon);
	/* load window icon pixbuf */
	kiba_winlist_icon_update_surface (icon);
	
	winlist->winlist_icons = g_list_append (winlist->winlist_icons, icon);
	
	if (icon->winlist_member == 1)
		kiba_winlist_object_new (icon, plugin);
	
	/* add window signal handler */
	g_signal_connect (G_OBJECT(icon->window), "icon-changed",
		  G_CALLBACK(icon_changed_callback), icon);
	g_signal_connect (G_OBJECT(icon->window), "state-changed",
		  G_CALLBACK(state_changed_callback), icon);
	g_signal_connect (G_OBJECT(icon->window), "name-changed",
		  G_CALLBACK(name_changed_callback), icon);
	
	return icon;
}

/**
Plugin Functions
**/

static void plugin_unload (KibaPlugin *plugin) {
	KibaWinlist *winlist;
	GList *l;
	
	winlist = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(KIBA_SCREEN_DEFAULT),
		  G_CALLBACK(window_opened_callback), plugin);
	g_signal_handlers_disconnect_by_func (G_OBJECT(KIBA_SCREEN_DEFAULT),
		  G_CALLBACK(window_closed_callback), plugin);
	g_signal_handlers_disconnect_by_func (G_OBJECT(KIBA_SCREEN_DEFAULT),
		  G_CALLBACK(withdrawn_toggled_callback), plugin);
	g_signal_handlers_disconnect_by_func (G_OBJECT(KIBA_SCREEN_DEFAULT),
		  G_CALLBACK(active_window_changed_callback), winlist);
	kiba_prefs_remove_callback_by_data (winlist);
	
	for (l = winlist->winlist_icons; l != NULL; l = winlist->winlist_icons) {
		KibaWinlistIcon *icon = l->data;
		
		winlist->winlist_icons = g_list_remove (winlist->winlist_icons, l->data);
		
		if (icon->title)
			g_free (icon->title);
		
		if (icon->window) {
			g_signal_handlers_disconnect_by_func (G_OBJECT(icon->window),
				G_CALLBACK(icon_changed_callback), icon);
			g_signal_handlers_disconnect_by_func (G_OBJECT(icon->window),
				G_CALLBACK(state_changed_callback), icon);
			g_signal_handlers_disconnect_by_func (G_OBJECT(icon->window),
				G_CALLBACK(name_changed_callback), icon);
		}
		
		if (icon->winlist_member == 1) {
			kiba_object_destroy (KIBA_OBJECT(icon->object));
		}
		else {
			if (icon->surface_data->surface)
				cairo_surface_destroy (icon->surface_data->surface);
			g_free (icon->surface_data);
			icon->surface_data = NULL;
		}
		
		g_free (icon);
	}
	g_list_free (winlist->winlist_icons);
	
	g_free (winlist);
}

static gboolean plugin_load (KibaPlugin *plugin) {
	KibaWinlist *winlist;

	winlist = g_new0 (KibaWinlist, 1);
	
	g_object_set_data (G_OBJECT(plugin),
		PLUGIN_NAME, (gpointer) winlist);
	
	/* Pref Callbacks */
	kiba_prefs_add_callback ("dock_object_size",
		dock_object_size_cb, winlist, TRUE);
	kiba_prefs_add_callback ("winlist_size_factor",
		object_size_factor_cb, winlist, TRUE);
	kiba_prefs_add_callback ("winlist_alpha",
		object_alpha_cb, winlist, FALSE);
	kiba_prefs_add_callback ("winlist_skip_app",
		skip_app_cb, winlist, TRUE);
	
	/* add a icon for each window */
	kiba_screen_for_each_window (KIBA_SCREEN_DEFAULT,
		(GFunc) kiba_winlist_icon_new, (gpointer) plugin);
	
	/* add screen signal handler */
	g_signal_connect (G_OBJECT(KIBA_SCREEN_DEFAULT), "window-opened",
		  G_CALLBACK(window_opened_callback), plugin);
	g_signal_connect (G_OBJECT(KIBA_SCREEN_DEFAULT), "window-closed",
		  G_CALLBACK(window_closed_callback), plugin);
	g_signal_connect (G_OBJECT(KIBA_SCREEN_DEFAULT), "withdrawn-toggled",
		  G_CALLBACK(withdrawn_toggled_callback), plugin);
	g_signal_connect (G_OBJECT(KIBA_SCREEN_DEFAULT), "active-window-changed",
		  G_CALLBACK(active_window_changed_callback), winlist);
	
	return TRUE;
}

static PrefMember winlist_options[] = {
  { "winlist_alpha", "0.9", KIBA_DOUBLE },
  { "winlist_size_factor", "1.0", KIBA_DOUBLE },
  { "winlist_skip_app", "[cairo-clock]", KIBA_LIST }
};

static KibaPluginVtable vtable = {
	PLUGIN_NAME,			/* id/name        */
	"Winlist Plugin for Kiba-Dock",	/* summary        */
	NULL,				/* description    */	

	"Daniel Bessler <danielb@kiba-dock.org>",
					/* author         */
	"http://www.kiba-dock.org",	/* homepage       */

	plugin_load,			/* load            */
	plugin_unload,			/* unload          */
};

KIBA_INIT_PLUGIN (vtable, winlist_options);

