/*
 *  Copyright (C) 2007  Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <libintl.h>
#include <math.h>
#include <gtk/gtk.h>

#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-fixed-object.h>
#include <kiba-object.h>
#include <kiba-plugin.h>
#include <kiba-prefs.h>
#include <kiba-debug.h>

#define _(string)   gettext (string)
#define PLUGIN_NAME "zoom"

#define INFLUENCE 150.0f
#define MAGNITUDE_MAX(zoom) zoom->factor - 1.0f

typedef struct _KibaZoom       KibaZoom;
typedef struct _KibaZoomObject KibaZoomObject;

typedef enum {
	KIBA_ZOOM_SIMPLE = 0,
	KIBA_ZOOM_PARABOLIC
}KibaZoomMode;

struct _KibaZoomObject {
	KibaZoom *zoom;
	GtkWidget *widget;
	float factor;
	gint zoom_id;
	KibaAllocation alloc;
};

struct _KibaZoom {
	GList *objects;
	
	KibaZoomMode mode;
	
	float magnitude;
	float influence;
	
	gint shrink_id;
	gint grow_id;
	
	float factor;
	float step;
	
	float mouse_x;
	float mouse_y;
	
	gboolean entered;
	gboolean dockPressed;
};

static gboolean dock_motion_cb  (GtkWidget        *widget,
				 GdkEventCrossing *event,
				 gpointer          data);

/**
Helper Functions
**/

static KibaZoomObject* kiba_zoom_get_object_for_widget (KibaZoom  *zoom,
							GtkWidget *widget) {
	GList *l;
	
	for (l = zoom->objects; l != NULL; l = l->next) {
		KibaZoomObject *object = l->data;
		
		if (object->widget == widget)
			return object;
	}
	
	return NULL;
}

/** simple zooming **/

static gint scale_down_object (gpointer data) {
	KibaZoomObject *object = data;
	KibaZoom *zoom = object->zoom;
	
// 	if (GET_FLAG(KIBA_OBJECT(object->widget)->state, KIBA_OBJECT_DRAGGING))
// 		return TRUE;
	
	kiba_object_queue_draw (KIBA_OBJECT(object->widget));
	
	object->factor -= zoom->step;
	
	if (object->factor <= 1.0) {
		object->factor = 1.0;
		kiba_object_update_scale_factor (KIBA_OBJECT(object->widget));
		object->zoom_id = 0;
		
		return FALSE;
	}
	
	kiba_object_update_scale_factor (KIBA_OBJECT(object->widget));
	
	return TRUE;
}

static gint scale_up_object (gpointer data) {
	KibaZoomObject *object = data;
	KibaZoom *zoom = object->zoom;
	
	if (GET_FLAG(KIBA_OBJECT(object->widget)->state, KIBA_OBJECT_DRAGGING))
		return TRUE;
	
	object->factor += zoom->step;
	
	if (object->factor >= zoom->factor) {
		object->factor = zoom->factor;
		kiba_object_update_scale_factor (KIBA_OBJECT(object->widget));
		kiba_object_queue_draw (KIBA_OBJECT(object->widget));
		object->zoom_id = 0;
		
		return FALSE;
	}
	
	kiba_object_queue_draw (KIBA_OBJECT(object->widget));
	kiba_object_update_scale_factor (KIBA_OBJECT(object->widget));
	
	return TRUE;
}

/**
parabolic zooming

note: algo from gnome-dock
**/

static gfloat
scaleIntegral (gfloat delta, gfloat fInfluence, gfloat fMagnitude)
{
	gfloat rate = fInfluence / M_PI_2;
	gfloat expansion = fMagnitude * rate;
	if (delta <= -fInfluence)
		return delta - expansion;
	if (delta >= +fInfluence)
		return delta + expansion;
	return delta + expansion * sin (delta / rate);
}

static gfloat
scale (gfloat delta, gfloat fInfluence, gfloat fMagnitude)
{
	gfloat rate = fInfluence / M_PI_2;
	if (delta <= -fInfluence)
		return 1.0;
	if (delta >= +fInfluence)
		return 1.0;
	return 1.0 + fMagnitude * cosf (delta / rate);
}

static gfloat
scaledPosition (gfloat fX, gfloat fMouseX, gfloat fInfluence, gfloat fMagnitude)
{
	return fMouseX + scaleIntegral (fX - fMouseX, fInfluence, fMagnitude);
}

static gint scale_down (gpointer data) {
	KibaZoom *zoom = data;
	
// 	if (kiba_dock_object_drag_active(kiba_dock_get ())) {
// 		return TRUE;
// 	}
	
	if (zoom->entered && !zoom->dockPressed) {
		zoom->shrink_id = 0;
		return FALSE;
	}
	
	if (zoom->magnitude > 0.01f)
		zoom->magnitude *= 0.8f;
	else
		zoom->magnitude = 0.0f;
	
	gtk_widget_queue_resize_no_redraw(GTK_WIDGET(kiba_dock_get ()));
	
	if (zoom->magnitude == 0.0f) {
		zoom->shrink_id = 0;
		return FALSE;
	}
	
	return TRUE;
}

static void __allocate_object_size (KibaDock *dock,
			      gpointer  data) {
	KibaZoom *zoom = data;
	gint border_width = gtk_container_get_border_width(GTK_CONTAINER(dock));
	gfloat fEnd = 0.0;
	gfloat fAdjustment = 0.0;
	gfloat fGap;
	gfloat fStart;
	gfloat fLastX;
	GList *children;
	GList *l;
	const KibaLayoutData *d;
	
	d = &kiba_layout_data[dock->d.position];
	
	fGap = (float) dock->d.object_space;
	
	for (l = zoom->objects; l != NULL; l = l->next) {
		KibaZoomObject *object = l->data;
		gfloat destroy_factor;
		gfloat scale_x;
		gfloat scale_y;
		GList *li;
		GtkWidget *child;
		
		child = gtk_bin_get_child(GTK_BIN(object->widget));
		
		destroy_factor = KIBA_OBJECT(object->widget)->destroy_factor;
		scale_x = KIBA_OBJECT(object->widget)->scale_x * destroy_factor;
		scale_y = KIBA_OBJECT(object->widget)->scale_y * destroy_factor;
		
		for (li = KIBA_OBJECT(object->widget)->custom_scale_funcs; li != NULL; li = li->next) {
			KibaObjectCustomScaleData *scale_data = li->data;
			float func_scale_x = 1.0;
			float func_scale_y = 1.0;
			
			if (scale_data->data == object)
				continue;
			if (!scale_data->scale_widget)
				continue;
			
			scale_data->scale_func (KIBA_OBJECT(object->widget), &func_scale_x, &func_scale_y, scale_data->data);
			
			scale_x *= func_scale_x;
			scale_y *= func_scale_y;
		}
		
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			if (child)
				fEnd += fGap + child->allocation.width;
			else
				fEnd += fGap + (KIBA_OBJECT(object->widget)->size.width * scale_x);
		} else {
			if (child)
				fEnd += fGap + child->allocation.height;
			else
				fEnd += fGap + (KIBA_OBJECT(object->widget)->size.height * scale_x);
		}
	}
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
		fStart = (GTK_WIDGET(dock)->parent->allocation.width - (int)fEnd)/2 + dock->d.offset;
	else
		fStart = (GTK_WIDGET(dock)->parent->allocation.height - (int)fEnd)/2 + dock->d.offset;
	fLastX = fStart;
	
	fEnd += fStart - fGap;
	
	children = gtk_container_get_children(GTK_CONTAINER(dock));
	
	for (l = children; l != NULL; l = l->next) {
		GtkWidget *child = l->data;
		KibaZoomObject *object;
		gfloat fScale;
		gfloat fStartX;
		gfloat fX;
		gfloat fScaledX;
		gfloat destroy_factor;
		gfloat scale_x;
		gfloat scale_y;
		GList *li;
		GtkWidget *obj_child;
		
		object = kiba_zoom_get_object_for_widget (zoom, child);
		if (!object)
			continue;
		
		obj_child = gtk_bin_get_child(GTK_BIN(object->widget));
		
		destroy_factor = KIBA_OBJECT(child)->destroy_factor;
		scale_x = KIBA_OBJECT(child)->scale_x * destroy_factor;
		scale_y = KIBA_OBJECT(child)->scale_y * destroy_factor;
		
		for (li = KIBA_OBJECT(child)->custom_scale_funcs; li != NULL; li = li->next) {
			KibaObjectCustomScaleData *scale_data = li->data;
			float func_scale_x = 1.0;
			float func_scale_y = 1.0;
			
			if (scale_data->data == object)
				continue;
			if (!scale_data->scale_widget)
				continue;
			
			scale_data->scale_func (KIBA_OBJECT(child), &func_scale_x, &func_scale_y, scale_data->data);
			
			scale_x *= func_scale_x;
			scale_y *= func_scale_y;
		}
		
		if (obj_child) {
			object->alloc.width  = obj_child->allocation.width;
			object->alloc.height = obj_child->allocation.height;
		}
		else {
			object->alloc.width  = (KIBA_OBJECT(child)->size.width * scale_x);
			object->alloc.height = (KIBA_OBJECT(child)->size.height * scale_y);
		}
		
		fStartX = fLastX;
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			fLastX += object->alloc.width + fGap;
			
			fX = fStartX + object->alloc.width * 0.5; // center of Object
		
			fScaledX = scaledPosition (fX, zoom->mouse_x, zoom->influence, zoom->magnitude);
			fScale = scale (fX - zoom->mouse_x, zoom->influence, zoom->magnitude);
			
			if (!KIBA_IS_FIXED_OBJECT(object->widget)) {
				object->alloc.y = (float) (GTK_WIDGET(dock)->allocation.y +
					(d->valignment/2)*(GTK_WIDGET(dock)->allocation.height - object->alloc.height*fScale) -
					(d->valignment-1)*dock->d.object_margin);
			}
			else {
				object->alloc.y = (float) (GTK_WIDGET(dock)->allocation.y +
					(d->valignment/2)*
					(GTK_WIDGET(dock)->allocation.height - dock->d.height - border_width));
				object->alloc.height = dock->d.height;
			}
			
			object->alloc.x = fScaledX - (fX - fStartX) * fScale;
			object->factor = fScale;
			object->alloc.width  *= fScale;
			object->alloc.height *= fScale;
		} else {
			fLastX += object->alloc.height + fGap;
			
			fX = fStartX + object->alloc.height * 0.5; // center of Object
		
			fScaledX = scaledPosition (fX, zoom->mouse_y, zoom->influence, zoom->magnitude);
			fScale = scale (fX - zoom->mouse_y, zoom->influence, zoom->magnitude);
			
			if (!KIBA_IS_FIXED_OBJECT(object->widget)) {
				object->alloc.x = (float) (GTK_WIDGET(dock)->allocation.x +
					(d->halignment/2)*(GTK_WIDGET(dock)->allocation.width - object->alloc.width*fScale) -
					(d->halignment-1)*dock->d.object_margin);
			}
			else {
				object->alloc.x = (float) (GTK_WIDGET(dock)->allocation.x +
					(d->halignment/2)*
					(GTK_WIDGET(dock)->allocation.width - dock->d.height - border_width));
				object->alloc.width = dock->d.height;
			}
			
			object->alloc.y = fScaledX - (fX - fStartX) * fScale;
			object->factor = fScale;
			object->alloc.width  *= fScale;
			object->alloc.height *= fScale;
		}
	}
	
	if (g_list_length(zoom->objects) > 1) {
		GtkWidget *last_widget;
		GtkWidget *first_widget;
		KibaZoomObject *first;
		KibaZoomObject *last;
		float fWidth;
		
		first_widget = children->data;
		first = kiba_zoom_get_object_for_widget (zoom, first_widget);
		l = g_list_last(children);
		last_widget = l->data;
		last = kiba_zoom_get_object_for_widget (zoom, last_widget);
		
		if (first && last) {
			if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
				if (zoom->mouse_x < fStart)
					fAdjustment = fStart - first->alloc.x;
				else if (zoom->mouse_x > fEnd)
					fAdjustment = fEnd - last->alloc.width - last->alloc.x;
			} else {
				if (zoom->mouse_y < fStart)
					fAdjustment = fStart - first->alloc.y;
				else if (zoom->mouse_x > fEnd)
					fAdjustment = fEnd - last->alloc.height - last->alloc.y;
			}
		}
	}
	
	for (l = zoom->objects; l != NULL; l = l->next) {
		KibaZoomObject *object = l->data;
		
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
			object->alloc.x += fAdjustment + border_width;
		else
			object->alloc.y += fAdjustment + border_width;
		
		if (object->alloc.y != KIBA_OBJECT(object->widget)->alloc.y ||
		    object->alloc.x != KIBA_OBJECT(object->widget)->alloc.x) {
			kiba_object_queue_draw(KIBA_OBJECT(object->widget));
			KIBA_OBJECT(object->widget)->alloc = object->alloc;
			kiba_object_update_scale_factor (KIBA_OBJECT(object->widget));
		}
	}
	
	/* update the widget size and queue draw */
	for (l = children; l != NULL; l = l->next) {
		GtkWidget *child = l->data;
		KibaZoomObject *object;
		
		object = kiba_zoom_get_object_for_widget (zoom, child);
		if (!object)
			continue;
		GtkAllocation allocation = (GtkAllocation) {
			(int) KIBA_OBJECT(object->widget)->alloc.x,
			(int) KIBA_OBJECT(object->widget)->alloc.y,
			(int) KIBA_OBJECT(object->widget)->alloc.width,
			(int) KIBA_OBJECT(object->widget)->alloc.height
		};
		gtk_widget_size_allocate(object->widget, &allocation);
	}
}

static void __allocate_dock_size (KibaWin *win,
                                  gpointer data) {
	GtkWidget *dock = KIBA_WIN_CHILD(win);
	GtkWidget *widget = GTK_WIDGET(win);
	KibaZoom *zoom = data;
	GtkRequisition req;
	GtkAllocation allocation;
	gint border_size;
	GList *children;
	const KibaLayoutData *d;
	
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	border_size = gtk_container_get_border_width(GTK_CONTAINER(dock));

	children = gtk_container_get_children (GTK_CONTAINER(dock));

/* the object size must be updated before the dock size is calculated */
__allocate_object_size(KIBA_DOCK(dock), data);
	
	allocation = dock->allocation;
	
	/* get the size request from the dock */
	gtk_widget_size_request (dock, &req);
	allocation.width = req.width;
	allocation.height = req.height;
	
	if (KIBA_DOCK(dock)->autohide.state == DOCK_UNHIDDEN) {
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			allocation.y = d->valignment*(GTK_WIDGET (widget)->allocation.height - allocation.height) / 2;
			if (allocation.y + allocation.height > GTK_WIDGET (widget)->allocation.height)
				allocation.y = GTK_WIDGET (widget)->allocation.height - allocation.height;
			else if (allocation.y <= 0)
				allocation.y = 1;
		} else {
			allocation.x = d->halignment*(GTK_WIDGET (widget)->allocation.width - allocation.width) / 2;
			if (allocation.x + allocation.width > GTK_WIDGET (widget)->allocation.width)
				allocation.x = GTK_WIDGET (widget)->allocation.width - allocation.width;
			else if (allocation.x <= 0)
				allocation.x = 1;
		}
	}
	
	if (children) {
		KibaObject *first = KIBA_OBJECT(children->data);
		KibaObject *last = KIBA_OBJECT(g_list_last(children)->data);
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			gint new_x;
			new_x =  (int) first->alloc.x - KIBA_DOCK(dock)->d.cap_size - border_size;
			allocation.x = new_x;
			allocation.width = (int) (last->alloc.x + last->alloc.width) - allocation.x + KIBA_DOCK(dock)->d.cap_size + border_size;
		} else {
			gint new_y;
			new_y =  (int) first->alloc.y - KIBA_DOCK(dock)->d.cap_size - border_size;
			allocation.y = new_y;
			allocation.height = (int) (last->alloc.y + last->alloc.height) - allocation.y + KIBA_DOCK(dock)->d.cap_size + border_size;
		}
	}
	
	gtk_widget_size_allocate (dock, &allocation);
}

static gint grow_up (gpointer data) {
	KibaZoom *zoom = data;
	gint iMouseX;
	gint iMouseY;
	
	if (kiba_dock_object_drag_active(kiba_dock_get ()) || zoom->dockPressed)
		return TRUE;
	
	gtk_widget_get_pointer (GTK_WIDGET(kiba_win_get_default()), &iMouseX, &iMouseY);
	
	zoom->mouse_x = (gfloat) iMouseX;
	zoom->mouse_y = (gfloat) iMouseY;
	
	if (zoom->magnitude < 0.01f)
		zoom->magnitude = 0.01f;
	else if (zoom->magnitude < zoom->factor - 1.05f)
		zoom->magnitude /= 0.75f;
	else
		zoom->magnitude = MAGNITUDE_MAX(zoom);
	
	gtk_widget_queue_resize_no_redraw(GTK_WIDGET(kiba_dock_get ()));
	
	if (zoom->magnitude == MAGNITUDE_MAX(zoom)) {
		zoom->grow_id = 0;
		return FALSE;
	}
	
	return TRUE;
}

static void __parabolic_enter_event_cb (KibaZoom *zoom) {
	gint mouseX;
	gint mouseY;
	
	zoom->entered = TRUE;
	
	if (zoom->grow_id || zoom->magnitude == MAGNITUDE_MAX(zoom))
		return;
	if (zoom->shrink_id) {
		g_source_remove (zoom->shrink_id);
		zoom->shrink_id = 0;
	}
	
	gtk_widget_get_pointer (GTK_WIDGET(kiba_win_get_default()), &mouseX, &mouseY);
	zoom->mouse_x = (gfloat) mouseX;
	zoom->mouse_y = (gfloat) mouseY;
	
	zoom->grow_id = g_timeout_add (20, grow_up, zoom);
}

static void __parabolic_leave_event_cb (KibaZoom *zoom) {
	gint mouseX;
	gint mouseY;
	
	zoom->entered = FALSE;
	
	if (zoom->magnitude == 0.0f)
		return;
	if (zoom->grow_id) {
		g_source_remove (zoom->grow_id);
		zoom->grow_id = 0;
	}
	
	gtk_widget_get_pointer (GTK_WIDGET(kiba_win_get_default()), &mouseX, &mouseY);
	zoom->mouse_x = (gfloat) mouseX;
	zoom->mouse_y = (gfloat) mouseY;
	
	zoom->shrink_id = g_timeout_add (40, scale_down, zoom);
}

/**
Signal Handler
**/

static void kiba_object_scale_func (KibaObject *object,
				    float     *scale_x,
				    float     *scale_y,
				    gpointer    data) {
	KibaZoomObject *zoom_object = data;
	
	*scale_x = zoom_object->factor;
	*scale_y = zoom_object->factor;
}

static void kiba_object_max_scale_func (KibaObject *object,
				        float     *scale_x,
				        float     *scale_y,
				        gpointer    data) {
	KibaZoomObject *zoom_object = data;
	
	*scale_x = zoom_object->zoom->factor;
	*scale_y = zoom_object->zoom->factor;
}

static void kiba_prefs_zoom_factor_cb (const char *option_id,
				       gpointer    value,
				       gpointer    data) {
	KibaZoom *zoom = data;
	
	zoom->factor = g_ascii_strtod((char*) value, NULL);
}

static void kiba_prefs_zoom_step_cb (const char *option_id,
				     gpointer    value,
				     gpointer    data) {
	KibaZoom *zoom = data;
	
	zoom->step = g_ascii_strtod((char*) value, NULL);
}

static void kiba_prefs_zoom_mode_cb (const char *option_id,
				     gpointer    value,
				     gpointer    data) {
	KibaDock *dock;
	KibaWin *win;
	KibaZoom *zoom = data;
	KibaZoomMode mode;
	GList *l;
	
	if (strcmp ((char*) value, "simple") == 0)
		mode = KIBA_ZOOM_SIMPLE;
	else
		mode = KIBA_ZOOM_PARABOLIC;
	
	if (mode == zoom->mode)
		return;
	
	dock = kiba_dock_get ();
	win = KIBA_WIN(GTK_WIDGET(dock)->parent);
	
	zoom->mode = mode;
	
	switch (zoom->mode) {
		case KIBA_ZOOM_SIMPLE:
			kiba_dock_remove_custom_children_position_func_by_data (dock, zoom);
			kiba_win_unset_children_allocate_func (win, __allocate_dock_size);
			
			g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
				G_CALLBACK(dock_motion_cb), zoom);
			
			for (l = zoom->objects; l != NULL; l = l->next) {
				KibaZoomObject *object = l->data;
				
				kiba_object_add_custom_scale_function (KIBA_OBJECT(object->widget),
					kiba_object_scale_func,
					kiba_object_max_scale_func,
					FALSE,
					object);
			}
			
			break;
		case KIBA_ZOOM_PARABOLIC:
			kiba_dock_set_custom_children_position_func (dock, __allocate_object_size, zoom);
			kiba_win_set_children_allocate_func (win, __allocate_dock_size, zoom);
			
			g_signal_connect (G_OBJECT(dock), "motion-notify-event",
				G_CALLBACK(dock_motion_cb), zoom);
			
			for (l = zoom->objects; l != NULL; l = l->next) {
				KibaZoomObject *object = l->data;
				
				kiba_object_add_custom_scale_function (KIBA_OBJECT(object->widget),
					kiba_object_scale_func,
					kiba_object_max_scale_func,
					TRUE,
					object);
			}
			
			break;
	}
}

static gboolean object_enter_event_cb  (GtkWidget        *widget,
					GdkEventCrossing *event,
					gpointer          data) {
	KibaZoomObject *object = data;
	switch (object->zoom->mode) {
		case KIBA_ZOOM_SIMPLE:
			if (object->zoom_id)
				g_source_remove (object->zoom_id);
			object->zoom_id = g_timeout_add (30, scale_up_object, object);
			break;
		case KIBA_ZOOM_PARABOLIC:
			__parabolic_enter_event_cb(object->zoom);
			break;
		default:
			break;
	}
	return FALSE;
}

static gboolean object_leave_event_cb  (GtkWidget        *widget,
					GdkEventCrossing *event,
					gpointer          data) {
	KibaZoomObject *object = data;
	switch (object->zoom->mode) {
		case KIBA_ZOOM_SIMPLE:
			if (object->zoom_id)
				g_source_remove (object->zoom_id);
			object->zoom_id = g_timeout_add (30, scale_down_object, object);
			break;
		case KIBA_ZOOM_PARABOLIC:
			__parabolic_leave_event_cb(object->zoom);
			break;
		default:
			break;
	}
	return FALSE;
}

static gboolean dock_enter_event_cb  (GtkWidget        *widget,
					GdkEventCrossing *event,
					gpointer          data) {
	KibaZoom *zoom = data;
	switch (zoom->mode) {
		case KIBA_ZOOM_PARABOLIC:
			__parabolic_enter_event_cb(zoom);
			break;
		default:
			break;
	}
	return FALSE;
}

static gboolean dock_leave_event_cb  (GtkWidget        *widget,
					GdkEventCrossing *event,
					gpointer          data) {
	KibaZoom *zoom = data;
	switch (zoom->mode) {
		case KIBA_ZOOM_PARABOLIC:
			__parabolic_leave_event_cb(zoom);
			break;
		default:
			break;
	}
	return FALSE;
}

static gboolean dock_press_event_cb (GtkWidget *dock,
                                     GdkEventButton *button,
                                     gpointer data) {
	KibaZoom *zoom = data;
	if (button->button == 3) return FALSE;
	zoom->dockPressed = TRUE;
	if (zoom->mode != KIBA_ZOOM_PARABOLIC)
		return FALSE;
	if (zoom->entered) {
		__parabolic_leave_event_cb(zoom);
		zoom->entered = TRUE;
	}
	return FALSE;
}

static gboolean dock_release_event_cb (GtkWidget *dock,
                                     GdkEventButton *button,
                                     gpointer data) {
	KibaZoom *zoom = data;
	zoom->dockPressed = FALSE;
	if (zoom->mode != KIBA_ZOOM_PARABOLIC)
		return FALSE;
	if (zoom->entered)
		__parabolic_enter_event_cb(zoom);
	return FALSE;
}

static void dock_object_add_cb (GtkContainer *container,
				GtkWidget    *widget,
				gpointer      data) {
	KibaZoom *zoom = data;
	KibaZoomObject *object;
	
	if (kiba_zoom_get_object_for_widget (zoom, widget) != NULL)
		return;
	
	object = g_new0 (KibaZoomObject, 1);
	object->widget = widget;
	object->factor = 1.0;
	object->zoom = zoom;
	object->zoom_id = 0;
	
	zoom->objects = g_list_append (zoom->objects, object);
	
	kiba_object_add_custom_scale_function (KIBA_OBJECT(widget),
		kiba_object_scale_func,
		kiba_object_max_scale_func,
		zoom->mode == KIBA_ZOOM_PARABOLIC,
		object);
	
	g_signal_connect (G_OBJECT(widget), "enter-notify-event",
		G_CALLBACK(object_enter_event_cb), object);
	g_signal_connect (G_OBJECT(widget), "leave-notify-event",
		G_CALLBACK(object_leave_event_cb), object);
}

static void dock_object_remove_cb (GtkContainer *container,
				   GtkWidget    *widget,
				   gpointer      data) {
	KibaZoom *zoom = data;
	KibaZoomObject *object;
	
	object = kiba_zoom_get_object_for_widget (zoom, widget);
	
	if (!object)
		return;
	
	if (object->zoom_id)
		g_source_remove (object->zoom_id);
	
	zoom->objects = g_list_remove (zoom->objects, object);
	
	g_free (object);
}

static gboolean dock_motion_cb  (GtkWidget        *widget,
				 GdkEventCrossing *event,
				 gpointer          data) {
	KibaZoom *zoom = data;
	gint iMouseX;
	gint iMouseY;
	
	if (zoom->magnitude != MAGNITUDE_MAX(zoom))
		return TRUE;
	
	if (zoom->shrink_id) {
		g_source_remove (zoom->shrink_id);
		zoom->shrink_id = 0;
	}
	
	gtk_widget_get_pointer (GTK_WIDGET(kiba_win_get_default()), &iMouseX, &iMouseY);
	zoom->mouse_x = (gfloat) iMouseX;
	zoom->mouse_y = (gfloat) iMouseY;
	
	gtk_widget_queue_resize_no_redraw(widget);
	
	return TRUE;
}

/**
Plugin Functions
**/

static void plugin_unload (KibaPlugin *plugin) {
	KibaZoom *zoom;
	KibaDock *dock;
	GList *l;
	
	dock = kiba_dock_get ();
	zoom = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	
	for (l = zoom->objects; l != NULL; l = zoom->objects) {
		KibaZoomObject *object = l->data;
		
		if (object->zoom_id)
			g_source_remove (object->zoom_id);
		
		if (object->factor > 1.0) {
			object->factor = 1.0;
			kiba_object_update_scale_factor (KIBA_OBJECT(object->widget));
		}
		
		g_signal_handlers_disconnect_by_func (G_OBJECT(object->widget),
			G_CALLBACK(object_enter_event_cb), object);
		g_signal_handlers_disconnect_by_func (G_OBJECT(object->widget),
			G_CALLBACK(object_leave_event_cb), object);
		
		kiba_object_remove_custom_scale_function_by_data (KIBA_OBJECT(object->widget),
			object);
		
		g_free (object);
		zoom->objects = g_list_remove (zoom->objects, l->data);
	}
	g_list_free (zoom->objects);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_object_add_cb), zoom);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_object_remove_cb), zoom);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_motion_cb), zoom);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_enter_event_cb), zoom);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_leave_event_cb), zoom);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_press_event_cb), zoom);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_release_event_cb), zoom);
	
	kiba_dock_remove_custom_children_position_func_by_data (dock, zoom);
	kiba_win_unset_children_allocate_func (KIBA_WIN(GTK_WIDGET(dock)->parent), __allocate_dock_size);
	
	g_free (zoom);
}

static gboolean plugin_load (KibaPlugin *plugin) {
	KibaZoom *zoom;
	KibaDock *dock;
	GtkContainer *container;
	GList *l;
	
	dock = kiba_dock_get ();
	container = GTK_CONTAINER(dock);
	
	zoom = g_new0(KibaZoom, 1);
	zoom->factor = 1.5;
	zoom->step = 0.05;
	zoom->objects = NULL;
	zoom->mode = KIBA_ZOOM_SIMPLE;
	zoom->influence = INFLUENCE;
	zoom->magnitude = 0.0f;
	zoom->shrink_id = 0;
	zoom->grow_id = 0;
	zoom->entered = FALSE;
	zoom->dockPressed = FALSE;
	
	kiba_prefs_add_callback ("zoom_factor",
		kiba_prefs_zoom_factor_cb, zoom, TRUE);
	kiba_prefs_add_callback ("zoom_step",
		kiba_prefs_zoom_step_cb, zoom, TRUE);
	kiba_prefs_add_callback ("zoom_mode",
		kiba_prefs_zoom_mode_cb, zoom, TRUE);
	
	g_signal_connect (G_OBJECT(dock), "remove",
		G_CALLBACK(dock_object_remove_cb), zoom);
	g_signal_connect_after (G_OBJECT(dock), "add",
		G_CALLBACK(dock_object_add_cb), zoom);
	g_signal_connect (G_OBJECT(dock), "enter-notify-event",
		G_CALLBACK(dock_enter_event_cb), zoom);
	g_signal_connect (G_OBJECT(dock), "leave-notify-event",
		G_CALLBACK(dock_leave_event_cb), zoom);
	g_signal_connect (G_OBJECT(dock), "button-press-event",
		G_CALLBACK(dock_press_event_cb), zoom);
	g_signal_connect (G_OBJECT(dock), "button-release-event",
		G_CALLBACK(dock_release_event_cb), zoom);
	for (l = gtk_container_get_children(container); l != NULL; l = l->next) {
		GtkWidget *child = l->data;
		
		dock_object_add_cb (container, child, zoom);
	}
	
	g_object_set_data (G_OBJECT(plugin),
		PLUGIN_NAME, (gpointer) zoom);
	
	return TRUE;
}

static PrefMember zoom_options[] = {
    { "zoom_factor", "1.5", KIBA_DOUBLE },
    { "zoom_step", "0.05", KIBA_DOUBLE },
    { "zoom_mode", "simple", KIBA_STRING },
};

static KibaPluginVtable vtable = {
	PLUGIN_NAME,			/* id/name        */
	"Zoom Plugin for Kiba-Dock",	/* summary        */
	NULL,				/* description    */	

	"Daniel Bessler <danielb@kiba-dock.org>",
					/* author         */
	"http://www.kiba-dock.org",	/* homepage       */

	plugin_load,			/* load            */
	plugin_unload,			/* unload          */
};

KIBA_INIT_PLUGIN (vtable, zoom_options);
